import datetime

import pandas as pd
import pytest

from ecl2df import EclFiles, compdat

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed",
        allow_module_level=True,
    )

WELOPEN_CASES = [
    # WELOPEN SHUT closes both well and connections
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-shut",
    ),
    # WELOPEN SHUT with explicit defaults, still closes both well and connections
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 5* /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-shut-explicit-defaults",
    ),
    # Test the zero handling in WELOPEN (zero value means apply to all)
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 0 0 0 0 0 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="zero-values",
    ),
    # Test the defaults handling in WELOPEN (default values are negative)
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' -1 -1 -1 -1 -1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="negative-values",
    ),
    # When item 3-7 are defaulted, the action applies to the well and not the
    # connections. Thus they should be kept open.
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'STOP' /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "OPEN"],
            ],
        ),
        id="welopen-stop-on-well",
    ),
    # Both 1*, 0 and -1 are default values and give the same result:
    # When all coordinates are defaulted, the connections are left OPEN
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
     'OP2' 1 1 1 1 'OPEN' /
     'OP3' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'STOP' 1* 1* 1* /
     'OP2' 'STOP' 0  0  0  /
     'OP3' 'STOP' -1 -1 -1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "OPEN"],
                [datetime.date(2000, 1, 1), "OP2", 1, 1, 1, 1, "OPEN"],
                [datetime.date(2000, 1, 1), "OP3", 1, 1, 1, 1, "OPEN"],
            ],
        ),
        id="welopen-stop-on-well-explicit-defaults",
    ),
    # In this test, the well connection is first SHUT, but then
    # actually opened by applying STOP on the well. The well is still
    # closed but the connection is open. Note that applying STOP to
    # the connection instead of the well ('OP1' 'STOP' 1 1 1 /) would
    # leave the connection still shut. This behavior has been tested
    # in the simulator.
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' /
    /
    DATES
     1 FEB 2000 /
    /
    WELOPEN
     'OP1' 'STOP' /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 2, 1), "OP1", 1, 1, 1, 1, "OPEN"],
            ],
        ),
        id="welopen-shut-then-stop-on-well",
    ),
    # Closes a connection with specified I, J, K
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 1 1 1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-shut-on-connection",
    ),
    # Test when one coordinate is defaulted. Shuts all connections in the
    # well with J==1 and K==1
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
     'OP1' 2 1 1 1 'OPEN' /
     'OP1' 1 1 2 2 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 0 1 1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "OP1", 2, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-with-defaulted-I-coordinate",
    ),
    # Test that all combinations of two defaulted coordinates are working
    # And that defaulting with *, 0 and -1 is treated the same
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 2 'OPEN' /
     'OP2' 1 1 1 1 'OPEN' /
     'OP2' 2 2 2 2 'OPEN' /
     'OP3' 1 1 1 1 'OPEN' /
     'OP3' 2 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 2*    1 /
     'OP2' 'SHUT' 0  1  0 /
     'OP3' 'SHUT' 1 -1 -1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP2", 2, 2, 2, 2, "OPEN"],
                [datetime.date(2000, 1, 1), "OP3", 2, 1, 1, 1, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "OP2", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "OP3", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-combinations-of-defaulted-coordinates",
    ),
    # Both wilcard well name and K coordinate defaulted
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1'  1 1 1 1 'OPEN' /
     'OP2'  1 1 2 2 'OPEN' /
     'PROD' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP*'  'SHUT' 1 1 0 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "PROD", 1, 1, 1, 1, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "OP2", 1, 1, 2, 2, "SHUT"],
            ],
        ),
        id="both-wildcard-wellname-and-defaulted-coordinates",
    ),
    # Compdat changing with time. The WELOPEN statement is only acting on
    # connections that have been defined at that date. In this test, the
    # connections with I==1 and I==2 will be SHUT, but not I==3
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1'  1 1 1 1 'OPEN' /
    /
    DATES
     1 FEB 2000 /
    /
    COMPDAT
     'OP1'  2 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1'  'SHUT' 0 1 1 /
    /
    DATES
     1 MAR 2000 /
    /
    COMPDAT
     'OP1'  3 1 1 1 'OPEN' /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "OPEN"],
                [datetime.date(2000, 2, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 2, 1), "OP1", 2, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 3, 1), "OP1", 3, 1, 1, 1, "OPEN"],
            ],
        ),
        id="welopen-defaults-compdat-changing-with-time",
    ),
    # Welopen defaults with START instead of DATES in the first timestep
    pytest.param(
        """
    START
     1 JAN 2000 /
    /
    COMPDAT
     'OP1'  1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1'  'SHUT' 0 1 1 /
    /

    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-defaults-start",
    ),
    # No dates at all
    pytest.param(
        """
    COMPDAT
     'OP1'  1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1'  'SHUT' 0 1 1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [None, "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-defaults-no-dates",
    ),
    # No start date, then a date later
    pytest.param(
        """
    COMPDAT
     'OP1'  1 1 1 1 'OPEN' /
    /
    DATES
     1 JAN 2000 /
    /
    WELOPEN
     'OP1'  'SHUT' 0 1 1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [None, "OP1", 1, 1, 1, 1, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-defaults-no-start-date",
    ),
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1'  1 1 1 2 'OPEN' /
    /
    WELOPEN
     'OP1'  'SHUT' 0 0 3 /
    /
    """,
        None,
        id="no-connections-matching-welopen-defaults",
        marks=pytest.mark.xfail(
            raises=ValueError,
            match="No connections are matching WELOPEN keyword",
        ),
    ),
    # Defaulted COMPLUMPs in WELOPEN not supported
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1'  1 1 1 2 'OPEN' /
    /
    COMPLUMP
     'OP1' 1 1 1 1 1 /
     'OP1' 1 1 1 1 2 /
    /

    WELOPEN
     'OP1'  'SHUT' 3* 1 0 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "SHUT"],
            ],
        ),
        id="defaulted-complump-in-welopen-not-supported",
        marks=pytest.mark.xfail(
            raises=ValueError,
            match="Zeros for C1/C2 is not implemented",
        ),
    ),
    # STOP on connection is the same as SHUT
    # (note that STOP on well gives OPEN connections)
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'STOP' 1 1 1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
            ],
        ),
        id="welopen-stop-on-connection-is-shut",
    ),
    # POPN is the same as OPEN
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'POPN' 1 1 1 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "OPEN"],
            ],
        ),
        id="welopen-popn-on-connection-is-open",
    ),
    # WELOPEN refers to a lumped connection, but COMPLUMP is missing
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 1 1 1 1 1 /
    /
    """,
        None,
        id="complump_missing",
        marks=pytest.mark.xfail(raises=ValueError),
    ),
    # Well is missing
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP2' 'SHUT' 1 1 1 /
    /
    """,
        None,
        id="operating-on-unknown-well",
        marks=pytest.mark.xfail(raises=ValueError),
    ),
    # Test J slicing
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 3 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 1 1 2  /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 3, 3, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "SHUT"],
            ],
        ),
        id="j-slicing",
    ),
    # Test multiple connections to the same cell
    pytest.param(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
     'OP2' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 0 0 0 /
     'OP2' 'OPEN' 0 0 0 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "OP2", 1, 1, 1, 1, "OPEN"],
            ],
        ),
        id="multiple-connnections-same-cell",
    ),
    # Test multiple time steps
    pytest.param(
        """
    DATES
     1 MAY 2001 /
    /

    COMPDAT
     'OP1' 33 110 31 31 'OPEN'  /
    /

    WELOPEN
     'OP1' 'SHUT' 0 0 0 /
    /

    TSTEP
     1 /

    COMPDAT
     'OP1' 34 111 32 32 'OPEN' /
    /

    TSTEP
     2 3 /

    COMPDAT
     'OP1' 35 111 33 33 'SHUT' /
    /
    """,
        pd.DataFrame(
            {
                "WELL": {0: "OP1", 1: "OP1", 2: "OP1"},
                "I": {0: 33, 1: 34, 2: 35},
                "J": {0: 110, 1: 111, 2: 111},
                "K1": {0: 31, 1: 32, 2: 33},
                "K2": {0: 31, 1: 32, 2: 33},
                "OP/SH": {0: "SHUT", 1: "OPEN", 2: "SHUT"},
                "DATE": {
                    0: datetime.date(2001, 5, 1),
                    1: datetime.date(2001, 5, 2),
                    2: datetime.date(2001, 5, 7),
                },
            }
        ),
        id="multiple-time-steps",
    ),
    pytest.param(
        """
    DATES
     1 MAY 2001 /
    /

    COMPDAT
     'OP1' 33 110 31 31 'OPEN'  /
    /

    WELOPEN
     'OP1' 'OPEN' 0 0 0/
    /

    TSTEP
     1 /

    COMPDAT
     'OP1' 34 111 32 32 'OPEN' /
    /

    TSTEP
     2 3 /

    COMPDAT
     'OP1' 35 111 33 33 'SHUT' /
    /
    """,
        pd.DataFrame(
            {
                "WELL": {0: "OP1", 1: "OP1", 2: "OP1"},
                "I": {0: 33, 1: 34, 2: 35},
                "J": {0: 110, 1: 111, 2: 111},
                "K1": {0: 31, 1: 32, 2: 33},
                "K2": {0: 31, 1: 32, 2: 33},
                "OP/SH": {0: "OPEN", 1: "OPEN", 2: "SHUT"},
                "DATE": {
                    0: datetime.date(2001, 5, 1),
                    1: datetime.date(2001, 5, 2),
                    2: datetime.date(2001, 5, 7),
                },
            }
        ),
        id="more-time-steps",
    ),
    pytest.param(
        """
    DATES
     1 MAY 2001 /
    /

    COMPDAT
     'OP1' 33 110 31 31 'OPEN'  /
     'OP2' 66 110 31 31 'OPEN'  /
    /

    WELOPEN
     'OP2' 'OPEN' 0 0 0/
    /

    DATES
     2 MAY 2001 /
    /

    COMPDAT
     'OP1' 34 111 32 32 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' 0 0 0 /
    /

    DATES
     3 MAY 2001 /
    /

    WELOPEN
     'OP1' 'OPEN' 0 0 0 /
     'OP2' 'SHUT' 0 0 0 /
    /
    """,
        pd.DataFrame(
            {
                "WELL": {
                    0: "OP1",
                    1: "OP2",
                    2: "OP1",
                    3: "OP1",
                    4: "OP1",
                    5: "OP1",
                    6: "OP2",
                },
                "I": {0: 33, 1: 66, 2: 33, 3: 34, 4: 33, 5: 34, 6: 66},
                "J": {0: 110, 1: 110, 2: 110, 3: 111, 4: 110, 5: 111, 6: 110},
                "K1": {0: 31, 1: 31, 2: 31, 3: 32, 4: 31, 5: 32, 6: 31},
                "K2": {0: 31, 1: 31, 2: 31, 3: 32, 4: 31, 5: 32, 6: 31},
                "OP/SH": {
                    0: "OPEN",
                    1: "OPEN",
                    2: "SHUT",
                    3: "SHUT",
                    4: "OPEN",
                    5: "OPEN",
                    6: "SHUT",
                },
                "DATE": {
                    0: datetime.date(2001, 5, 1),
                    1: datetime.date(2001, 5, 1),
                    2: datetime.date(2001, 5, 2),
                    3: datetime.date(2001, 5, 2),
                    4: datetime.date(2001, 5, 3),
                    5: datetime.date(2001, 5, 3),
                    6: datetime.date(2001, 5, 3),
                },
            }
        ),
        id="date-stepping",
    ),
    pytest.param(
        """
    DATES
     1 MAY 2001 /
    /

    COMPDAT
     'OP1' 33 110 1 2 'OPEN'  /
    /

    WELOPEN
     'OP1' 'SHUT' 33 110 1 /
    /

    DATES
     2 MAY 2001 /
    /

    WELOPEN
     'OP1' 'SHUT' 33 110 2 /
    /

    DATES
     3 MAY 2001 /
    /

    WELOPEN
     'OP1' 'OPEN' 0 0 0 /
    /
    """,
        pd.DataFrame(
            {
                "WELL": {0: "OP1", 1: "OP1", 2: "OP1", 3: "OP1", 4: "OP1"},
                "I": {0: 33, 1: 33, 2: 33, 3: 33, 4: 33},
                "J": {0: 110, 1: 110, 2: 110, 3: 110, 4: 110},
                "K1": {0: 2, 1: 1, 2: 2, 3: 1, 4: 2},
                "K2": {0: 2, 1: 1, 2: 2, 3: 1, 4: 2},
                "OP/SH": {0: "OPEN", 1: "SHUT", 2: "SHUT", 3: "OPEN", 4: "OPEN"},
                "DATE": {
                    0: datetime.date(2001, 5, 1),
                    1: datetime.date(2001, 5, 1),
                    2: datetime.date(2001, 5, 2),
                    3: datetime.date(2001, 5, 3),
                    4: datetime.date(2001, 5, 3),
                },
            }
        ),
        id="more-date-stepping",
    ),
    pytest.param(
        """
    DATES
     1 MAY 2001 /
    /

    COMPDAT
     'OP1' 1 1 1 1 'OPEN'  /
    /

    DATES
     2 MAY 2001 /
    /

    WELOPEN
     'OP1' 'SHUT' /
    /

    COMPDAT
     'OP1' 1 1 1 1 'OPEN'  /
    /
    """,
        pd.DataFrame(
            {
                "WELL": {0: "OP1", 1: "OP1"},
                "I": {0: 1, 1: 1},
                "J": {0: 1, 1: 1},
                "K1": {0: 1, 1: 1},
                "K2": {0: 1, 1: 1},
                "OP/SH": {0: "OPEN", 1: "OPEN"},
                "DATE": {0: datetime.date(2001, 5, 1), 1: datetime.date(2001, 5, 2)},
            }
        ),
        id="test-xx1",
    ),
    pytest.param(
        """
    DATES
     1 MAY 2001 /
    /

    COMPDAT
     'OP1' 1 1 1 1 'OPEN'  /
    /

    WELOPEN
     'OP1' 'OPEN' 0 0 0 /
     'OP1' 'SHUT' 0 0 0 /
    /
    """,
        pd.DataFrame(
            {
                "WELL": {0: "OP1"},
                "I": {0: 1},
                "J": {0: 1},
                "K1": {0: 1},
                "K2": {0: 1},
                "OP/SH": {0: "SHUT"},
                "DATE": {0: datetime.date(2001, 5, 1)},
            }
        ),
        id="self-overwriting-records",
    ),
    pytest.param(
        """
    DATES
     1 MAY 2001 /
    /

    COMPDAT
     'OP1' 1 1 1 2 'SHUT'  /
    /

    WELOPEN
     'OP1' 'OPEN' 0 0 0 /
     'OP1' 'SHUT' 1 1 1 /
    /
    """,
        pd.DataFrame(
            {
                "WELL": {0: "OP1", 1: "OP1"},
                "I": {0: 1, 1: 1},
                "J": {0: 1, 1: 1},
                "K1": {0: 2, 1: 1},
                "K2": {0: 2, 1: 1},
                "OP/SH": {0: "OPEN", 1: "SHUT"},
                "DATE": {0: datetime.date(2001, 5, 1), 1: datetime.date(2001, 5, 1)},
            }
        ),
        id="open-and-shut-slice-multiple-welopen",
    ),
    # Referencing multiple wells with wildcards
    # Wildcard structures are tested in test_common and need
    # not be tested here.
    pytest.param(
        """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
     'B_1H' 1 1 1 1 'OPEN' /
     'B_2H' 2 2 2 2 'OPEN' /
     'WI1' 3 3 3 3 'OPEN' /
    /
    WELOPEN
     'B*H' 'SHUT' 0 0 0 /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "WI1", 3, 3, 3, 3, "OPEN"],
                [datetime.date(2000, 1, 1), "B_1H", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "B_2H", 2, 2, 2, 2, "SHUT"],
            ],
        ),
        id="multiple-wells-via-wildcard",
    ),
    # Test wildcard in wellname. A well that also matches the well template
    # but is defined later, is not SHUT
    pytest.param(
        """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
     'OP2' 2 2 2 2 'OPEN' /
     'WI1' 3 3 3 3 'OPEN' /
    /
    WELOPEN
     'OP*' 'SHUT' 0 0 0 /
    /
    DATES
      1 FEB 2000 /
    /
    COMPDAT
      'OP3' 4 4 4 4 'OPEN' /
    /
    """,
        pd.DataFrame(
            columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
            data=[
                [datetime.date(2000, 1, 1), "WI1", 3, 3, 3, 3, "OPEN"],
                [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                [datetime.date(2000, 1, 1), "OP2", 2, 2, 2, 2, "SHUT"],
                [datetime.date(2000, 2, 1), "OP3", 4, 4, 4, 4, "OPEN"],
            ],
        ),
        id="wildcards-do-not-apply-to-future-wells",
    ),
]


@pytest.mark.parametrize("test_input, expected", WELOPEN_CASES)
def test_welopen(test_input, expected):
    """Test with WELOPEN present"""
    deck = EclFiles.str2deck(test_input)
    compdf = compdat.deck2dfs(deck)["COMPDAT"]
    columns_to_check = ["WELL", "I", "J", "K1", "K2", "OP/SH", "DATE"]

    pd.testing.assert_frame_equal(compdf[columns_to_check], expected[columns_to_check])


@pytest.mark.parametrize(
    # It should only be necessary to test "NEW" actions in WLIST here, as the
    # transformation from ADD/DEL/MOV into NEW is tested in test_wlist
    "test_input, expected",
    [
        # Simplest case with one well in list
        (
            """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
      'OP1' 1 1 1 1 'OPEN' /
    /
    WLIST
      '*OP' NEW OP1 /
    /
    WELOPEN
      '*OP' 'SHUT' 0 0 0 /
    /
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                ],
            ),
        ),
        # WLIST for a different well
        pytest.param(
            """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
      'OP1' 1 1 1 1 'OPEN' /
    /
    WLIST
      '*OP' NEW OP2 /
    /
    WELOPEN
      '*OP' 'SHUT' 0 0 0 /
    /
    """,
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, match="A WELOPEN keyword is not acting"
            ),
        ),
        # Two wells.
        (
            """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
      'OP1' 1 1 1 1 'OPEN' /
      'OP2' 1 1 1 1 'OPEN' /
    /
    WLIST
      '*OP' NEW OP1 OP2/
    /
    WELOPEN
      -- Shut the wells immediately, overriding OPEN in COMPDAT
      '*OP' 'SHUT' 0 0 0 /
    /
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                    [datetime.date(2000, 1, 1), "OP2", 1, 1, 1, 1, "SHUT"],
                ],
            ),
        ),
        # Four wells, two lists.
        (
            """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
      'OP1' 1 1 1 1 'OPEN' /
      'OP2' 1 1 1 1 'OPEN' /
      'IN1' 2 1 1 1 'OPEN' /
      'IN2' 2 1 1 1 'OPEN' /
    /
    WELOPEN
      -- In ecl2df, the WELOPEN is allowed to be before WLIST
      '*OP' 'SHUT' 0 0 0 /
    /
    WLIST
      '*OP' NEW OP1 OP2 /
      '*IN' NEW IN1 IN2 /
    /
    DATES
      2 JAN 2000 /
    /
    WELOPEN
      '*IN' 'SHUT' 0 0 0 /
    /
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "IN1", 2, 1, 1, 1, "OPEN"],
                    [datetime.date(2000, 1, 1), "IN2", 2, 1, 1, 1, "OPEN"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                    [datetime.date(2000, 1, 1), "OP2", 1, 1, 1, 1, "SHUT"],
                    [datetime.date(2000, 1, 2), "IN1", 2, 1, 1, 1, "SHUT"],
                    [datetime.date(2000, 1, 2), "IN2", 2, 1, 1, 1, "SHUT"],
                ],
            ),
        ),
        # WLIST defined too late
        pytest.param(
            """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
      'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
      '*OP' 'SHUT' 0 0 0 /
    /
    DATES
      2 JAN 2000/
    /
    WLIST
      '*OP' NEW OP2 /
    /
    """,
            None,
            id="futurewlist",
            marks=pytest.mark.xfail(
                raises=ValueError, match="Well list OP not defined at 2000-01-01"
            ),
        ),
        pytest.param(
            # WELOPEN on non-existing well list name
            """
    DATES
      1 JAN 2000 /
    /
    COMPDAT
      'OP1' 1 1 1 1 'OPEN' /
    /
    WLIST
      '*OP' NEW OP1 /
    /
    WELOPEN
      '*OPS' 'SHUT' 0 0 0 /
    /
    """,
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, match="Well list OPS not defined at 2000-01-01"
            ),
        ),
        # WLIST that has been redefined
        pytest.param(
            """
    DATES
      1 JAN 1999/
    /
    WLIST
      '*OP' NEW OP9 /
    /
    DATES
      1 JAN 2000 /
    /
    COMPDAT
      'OP1' 1 1 1 1 'OPEN' /
    /
    WLIST
      '*OP' NEW OP1 /
    /
    WELOPEN
      '*OP' 'SHUT' 0 0 0 /
    /
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                ],
            ),
            id="redefined_wlist",
        ),
    ],
)
def test_welopen_wlist(test_input, expected):
    deck = EclFiles.str2deck(test_input)
    dfs = compdat.deck2dfs(deck)
    pd.testing.assert_frame_equal(dfs["COMPDAT"][expected.columns], expected)


def test_welopen_df():
    """Test that we can obtain WELOPEN information when it applies on well state,
    not on connections."""
    deck = EclFiles.str2deck(
        """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    WELOPEN
     'OP1' 'SHUT' / -- This is ignored for connections
    /
    """
    )
    print(compdat.deck2dfs(deck)["WELOPEN"])
    pd.testing.assert_frame_equal(
        compdat.deck2dfs(deck)["WELOPEN"],
        pd.DataFrame(
            columns=[
                "DATE",
                "WELL",
                "I",
                "J",
                "K",
                "C1",
                "C2",
                "STATUS",
                "KEYWORD_IDX",
            ],
            data=[
                [
                    datetime.date(2000, 1, 1),
                    "OP1",
                    None,
                    None,
                    None,
                    None,
                    None,
                    "SHUT",
                    2,
                ],
            ],
        ),
        check_like=True,
    )


@pytest.mark.parametrize(
    "test_input, expected",
    [
        # Simplest possible case
        (
            """
DATES
    1 JAN 2000 /
/
COMPDAT
    'OP1' 1 1 1 1 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 1 1 1 /
/
WELOPEN
    'OP1' 'SHUT' 3* 1 1 /
/
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                ],
            ),
        ),
        # Range of Ks in COMPDAT/COMPLUMP and multiple dates
        (
            """
DATES
    1 JAN 2000 /
/
COMPDAT
    'OP1' 1 1 1 3 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 1 2 1 /
/
DATES
    1 FEB 2000 /
/
WELOPEN
    'OP1' 'SHUT' 3* 1 1 /
/
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "OPEN"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "OPEN"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 3, 3, "OPEN"],
                    [datetime.date(2000, 2, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                    [datetime.date(2000, 2, 1), "OP1", 1, 1, 2, 2, "SHUT"],
                ],
            ),
        ),
        # Range of COMPLUMPs in WELOPEN
        (
            """
DATES
    1 JAN 2000 /
/
COMPDAT
    'OP1' 1 1 1 5 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 1 2 1 /
    'OP1' 1 1 3 4 2 /
    'OP1' 1 1 5 5 3 /
/
WELOPEN
    'OP1' 'SHUT' 3* 1 2 /
/
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 5, 5, "OPEN"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "SHUT"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 3, 3, "SHUT"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 4, 4, "SHUT"],
                ],
            ),
        ),
        # Test default handling in COMPLUMP. Default value 0 means all values
        # of this coordinate
        # This fails for now, but dataframe is indicating wanted behavior
        pytest.param(
            """
DATES
    1 JAN 2000 /
/
COMPDAT
    'OP1' 1 1 1 2 'OPEN' /
    'OP1' 2 1 1 1 'OPEN' /
/
COMPLUMP
    'OP1' 1 0 0 0 1 /
/
WELOPEN
    'OP1' 'SHUT' 3* 1 1 /
/
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "SHUT"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "OPEN"],
                ],
            ),
            id="complump_defaults",
            marks=pytest.mark.xfail(
                raises=ValueError,
                match="Defaulted COMPLUMP coordinates are not supported in ecl2df",
            ),
        ),
        pytest.param(
            # Fails when K2<K1 in COMPLUMP
            """
COMPDAT
    'OP1' 1 1 1 2 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 2 1 1 /
/
WELOPEN
    'OP1' 'SHUT' 3* 1 1 /
/
""",
            None,
            id="complump_K2<K1",
            marks=pytest.mark.xfail(
                raises=ValueError, match="K2 must be equal to or greater than K1"
            ),
        ),
        pytest.param(
            # Fails when only one completion number is defined in WELOPEN
            """
COMPDAT
    'OP1' 1 1 1 1 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 1 1 1 /
/
WELOPEN
    'OP1' 'SHUT' 3* 1 /
/
""",
            None,
            id="complump_missingcompletion_number",
            marks=pytest.mark.xfail(
                raises=ValueError,
                match=(
                    "Both or none of the completions numbers "
                    "in WELOPEN must be defined."
                ),
            ),
        ),
        pytest.param(
            # Fails when C2<C1 in WELOPEN
            """
COMPDAT
    'OP1' 1 1 1 2 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 1 1 1 /
    'OP1' 1 1 2 2 2 /
/
WELOPEN
    'OP1' 'SHUT' 3* 2 1 /
/
""",
            None,
            id="welopen_C2<C1",
            marks=pytest.mark.xfail(
                raises=ValueError, match="C2 must be equal or greater than C1"
            ),
        ),
        pytest.param(
            # Fails for negative numbers in COMPLUMP row
            """
COMPDAT
    'OP1' 1 1 1 1 'OPEN' /
/
COMPLUMP
    'OP1' -1 -1 -1 -1 1 /
/
WELOPEN
    'OP1' 'SHUT' 3* 1 1 /
/
""",
            None,
            id="complump_negativevalues",
            marks=pytest.mark.xfail(
                raises=ValueError,
                match="Negative values for COMPLUMP coordinates are not allowed",
            ),
        ),
        pytest.param(
            # Fails for negative completion numbers in welopen
            """
COMPDAT
    'OP1' 1 1 1 1 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 1 1 /
/
WELOPEN
    'OP1' 'SHUT' 3* -1 -1 /
/
""",
            None,
            id="welopen_negative_completionvalues",
            marks=pytest.mark.xfail(
                raises=ValueError,
                match="Negative values for C1/C2 is no allowed",
            ),
        ),
        pytest.param(
            # Fails for default completionvalues (zero) in WELOPEN
            """
COMPDAT
    'OP1' 1 1 1 1 'OPEN' /
/
COMPLUMP
    'OP1' 1 1 1 1 /
/
WELOPEN
    'OP1' 'SHUT' 3* 0 0 /
/
""",
            None,
            id="welopen_default_complumpvalues",
            marks=pytest.mark.xfail(
                raises=ValueError,
                match="Defaults (zero) for C1/C2 is not implemented",
            ),
        ),
        pytest.param(
            """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
    /
    COMPLUMP
     'OP1' 1 1 1 1 1 /
    /
    WELOPEN
     'OP1' 'SHUT' 1 1 1 1 1 /
    /
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                ],
            ),
            id="indices_and_complump_combined",
        ),
        pytest.param(
            """
    DATES
     1 JAN 2000 /
    /
    COMPDAT
     'OP1' 1 1 1 1 'OPEN' /
     'OP1' 1 1 2 2 'OPEN' /
    /
    COMPLUMP
     -- Assign completion number 1 and 2 to the two connections
     'OP1' 1 1 1 1 1 /
     'OP1' 1 1 2 2 2 /
    /
    WELOPEN
     -- This is ok, gives shut well:
     'OP1' 'SHUT' 1 1 1 1 1 /  -- must match both i,j,k and complump
     -- The following is ignored, because completion 2 is not at 1,1,1
     'OP1' 'SHUT' 1 1 1 2 2 /
    /
    """,
            pd.DataFrame(
                columns=["DATE", "WELL", "I", "J", "K1", "K2", "OP/SH"],
                data=[
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 2, 2, "OPEN"],
                    [datetime.date(2000, 1, 1), "OP1", 1, 1, 1, 1, "SHUT"],
                ],
            ),
            id="indices_and_complump_combined_2",
        ),
    ],
)
def test_welopen_complump(test_input, expected):
    """Test the welopen_complump functionality through Eclipse decks"""
    deck = EclFiles.str2deck(test_input)
    dfs = compdat.deck2dfs(deck)
    pd.testing.assert_frame_equal(dfs["COMPDAT"][expected.columns], expected)


@pytest.mark.parametrize(
    "welopen, complump, expected",
    [
        pytest.param([], [], [], id="empty_input_empty_output"),
        pytest.param(
            [{"FOO": "BAR"}],
            [],
            [{"FOO": "BAR"}],
            id="empty_complump_pass_through_welopen",
        ),
        pytest.param(
            [{"WELL": "OP1", "C1": None, "C2": None}],
            [],
            [{"WELL": "OP1", "C1": None, "C2": None}],
            id="pass_through_None_complumps",
        ),
        pytest.param(
            [{"WELL": "OP1", "DATE": datetime.date(2000, 1, 1), "C1": 1, "C2": 1}],
            [
                {
                    "WELL": "OP1",
                    "DATE": datetime.date(2000, 1, 1),
                    "STATUS": "OPEN",
                    "I": 2,
                    "J": 2,
                    "K1": 2,
                    "K2": 2,
                    "N": 1,
                }
            ],
            [
                {
                    "WELL": "OP1",
                    "DATE": datetime.date(2000, 1, 1),
                    "C1": None,
                    "C2": None,
                    "I": 2,
                    "J": 2,
                    "K": 2,
                }
            ],
            id="basic",
        ),
        pytest.param(
            [{"WELL": "OP1", "DATE": datetime.date(2000, 1, 1), "C1": 1, "C2": 1}],
            [
                {
                    "WELL": "OP1",
                    "DATE": datetime.date(2000, 1, 1),
                    "STATUS": "OPEN",
                    "I": 2,
                    "J": 2,
                    "K1": 2,
                    "K2": 3,
                    "N": 1,
                }
            ],
            [],
            id="k2_not_equal_kq_in_complump",
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_welopen_complump_direct(welopen, complump, expected):
    """Test the welopen_complump directly.

    For situations not that easy to reach via string input."""
    pd.testing.assert_frame_equal(
        compdat.expand_complump_in_welopen_df(
            pd.DataFrame(welopen), pd.DataFrame(complump)
        ),
        pd.DataFrame(expected),
        check_dtype=False,
    )


@pytest.mark.parametrize(
    "compdat_rows, welopen_rows, wlist_rows, complump_rows, expected_rows",
    [
        pytest.param([], [], [], [], [], id="emptyinput"),
        pytest.param(
            [{"WELL": "OP1"}],
            [],
            [{"ACTION": "ADD"}],
            [],
            [],
            id="non-expanded_wlist",
            # "The WLIST dataframe must be expanded through expand_wlist()"
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            [{"WELL": "OP1", "I": 1, "J": 1, "K1": 1, "K2": 1}],
            [{"WELL": "OP1", "STATUS": "SHUT", "I": 0, "J": 0, "K": 0}],
            [],
            [],
            [],
            id="no_keyword_idx",
            # Column KEYWORD_IDX must be present in compdat rows
            marks=pytest.mark.xfail(raises=KeyError),
        ),
        pytest.param(
            [
                {
                    "WELL": "OP1",
                    "I": 1,
                    "J": 1,
                    "K1": 1,
                    "K2": 1,
                    "KEYWORD_IDX": 1,
                    "DATE": None,
                }
            ],
            [
                {
                    "WELL": "OP1",
                    "STATUS": "SHUT",
                    "I": 0,
                    "J": 0,
                    "K": 0,
                    "KEYWORD_IDX": 2,
                    "DATE": None,
                }
            ],
            [],
            [],
            [
                {
                    "WELL": "OP1",
                    "I": 1,
                    "J": 1,
                    "K1": 1,
                    "K2": 1,
                    "KEYWORD_IDX": 2,
                    "DATE": None,
                    "OP/SH": "SHUT",
                }
            ],
            id="working_example",
        ),
        pytest.param(
            [
                {
                    "WELL": "OP1",
                    "I": 1,
                    "J": 1,
                    "K1": 1,
                    "K2": 1,
                    "KEYWORD_IDX": 1,
                    "DATE": None,
                }
            ],
            [
                {
                    "WELL": "OP1",
                    "STATUS": "SHUT",
                    "I": 1,  # Either all zero or none nonzero
                    "J": 0,
                    "K": 0,
                    "KEYWORD_IDX": 2,
                    "DATE": None,
                }
            ],
            [],
            [],
            [],
            id="invalid welopen",
            # " A WELOPEN keyword contains data that could not be parsed"
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_applywelopen(
    compdat_rows, welopen_rows, wlist_rows, complump_rows, expected_rows
):
    print(
        compdat.applywelopen(
            pd.DataFrame(compdat_rows),
            pd.DataFrame(welopen_rows),
            pd.DataFrame(wlist_rows),
            pd.DataFrame(complump_rows),
        ).to_dict(orient="records")
    )
    pd.testing.assert_frame_equal(
        compdat.applywelopen(
            pd.DataFrame(compdat_rows),
            pd.DataFrame(welopen_rows),
            pd.DataFrame(wlist_rows),
            pd.DataFrame(complump_rows),
        ),
        pd.DataFrame(expected_rows),
    )



"""Common functions for ecl2df modules"""

import argparse
import datetime
import inspect
import itertools
import json
import logging
import re
import shlex
import signal
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

import numpy as np
import pandas as pd
import pyarrow

try:
    # pylint: disable=unused-import
    import opm.io.deck  # lgtm [py/import-and-import-from]

    # This import is seemingly not used, but necessary for some attributes
    # to be included in DeckItem objects.
    from opm.io.deck import DeckKeyword  # noqa
except ImportError:
    # Allow parts of ecl2df to work without OPM:
    pass

from ecl2df import __version__

from .constants import MAGIC_STDOUT

# Parse named JSON files, this exposes a dict of dictionary describing the contents
# of supported Eclipse keyword data
OPMKEYWORDS: Dict[str, dict] = {}
for keyw in [
    "BRANPROP",
    "COMPDAT",
    "COMPLUMP",
    "COMPSEGS",
    "DENSITY",
    "EQLDIMS",
    "EQUIL",
    "FAULTS",
    "GRUPNET",
    "GRUPTREE",
    "NODEPROP",
    "PBVD",
    "PDVD",
    "PVDG",
    "PVDO",
    "PVTG",
    "PVTO",
    "PVTW",
    "ROCK",
    "RSVD",
    "RVVD",
    "SGFN",
    "SGWFN",
    "SGOF",
    "SLGOF",
    "SOF2",
    "SOF3",
    "SWFN",
    "SWOF",
    "TABDIMS",
    "WCONHIST",
    "WCONINJE",
    "WCONINJH",
    "WCONPROD",
    "WELOPEN",
    "WELSEGS",
    "WELSPECS",
    "WLIST",
    "WSEGAICD",
    "WSEGSICD",
    "WSEGVALV",
]:
    OPMKEYWORDS[keyw] = json.loads(
        (Path(__file__).parent / "opmkeywords" / keyw).read_text()
    )


SVG_COLOR_NAMES = [
    color.lower()
    for color in (
        (Path(__file__).parent / "svg_color_keyword_names.txt")
        .read_text(encoding="utf-8")
        .splitlines()
    )
]

logger: logging.Logger = logging.getLogger(__name__)


def write_dframe_stdout_file(
    dframe: Union[pd.DataFrame, pyarrow.Table],
    output: str,
    index: bool = False,
    caller_logger: Optional[logging.Logger] = None,
    logstr: Optional[str] = None,
) -> None:
    """Write a dataframe to either stdout or a file

    If output is the magic string "-", output is written
    to stdout.

    Arguments:
        dframe: Dataframe to write
        output: Filename or "-"
        index: Passed to to_csv()
        caller_logger: Used if not stdout
        logstr: Logged if not stdout.
    """
    if output == MAGIC_STDOUT:
        # Ignore pipe errors when writing to stdout:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        if isinstance(dframe, pd.DataFrame):
            dframe.to_csv(sys.stdout, index=index)
        else:
            raise SystemExit("Not possible to write arrow format to stdout")
    else:
        if caller_logger and isinstance(dframe, pd.DataFrame) and dframe.empty:
            caller_logger.warning("Empty dataframe being written to disk")
        if caller_logger and not logstr:
            caller_logger.info("Writing to file %s", str(output))
        elif caller_logger and logstr:
            caller_logger.info(logstr)
        if isinstance(dframe, pd.DataFrame):
            dframe.to_csv(output, index=index)
        else:
            pyarrow.feather.write_feather(dframe, dest=output)


def write_inc_stdout_file(string: str, outputfilename: str) -> None:
    """Write a string (typically an include file string) to stdout
    or to a named file"""
    if outputfilename == MAGIC_STDOUT:
        # Ignore pipe errors when writing to stdout:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        print(string)
    else:
        Path(outputfilename).write_text(string, encoding="utf-8")
        print(f"Wrote to {outputfilename}")


def parse_ecl_month(eclmonth: str) -> int:
    """Translate Eclipse month strings to integer months"""
    eclmonth2num = {
        "JAN": 1,
        "FEB": 2,
        "MAR": 3,
        "APR": 4,
        "MAY": 5,
        "JUN": 6,
        "JUL": 7,
        "JLY": 7,
        "AUG": 8,
        "SEP": 9,
        "OCT": 10,
        "NOV": 11,
        "DEC": 12,
    }
    return eclmonth2num[eclmonth]


def ecl_keyworddata_to_df(
    deck,
    keyword: str,
    renamer: Optional[Dict[str, Union[str, List[str]]]] = None,
    recordcountername: Optional[str] = None,
    emptyrecordcountername: Optional[str] = None,
) -> pd.DataFrame:
    """Extract data associated to an Eclipse keyword into a tabular form.

    Two modes of enumeration of tables in the keyworddata is supported, you
    will have to find out which one fits your particular keyword. Activate
    by setting recordcountername or emptyrecordcountername to some string, which
    will be the name of your enumeration, e.g. PVTNUM, EQLNUM or SATNUM.

    Arguments:
        deck: Parsed deck
        keyword: Name of the keyword for which to extract data.
        renamer: Mapping of names present in OPM json files for the
            keyword to desired column names in returned dataframe
        recordcountername: If present, an extra column is added with this name
            with consecutive rows enumerated from 1. Use this to assign
            EQLNUM or similar when this should be consecutive pr. row (not
            the case for all keywords).
        emptyrecordcountername: If supplied, an index is added to every parsed
            row based on how many empty records is encountered. For PVTO f.ex,
            this gives the PVTNUM indexing.
    """
    records = []  # list of dicts or dataframes

    record_counter = 1
    emptyrecord_counter = 1
    for deckrecord in deck[keyword]:
        if str(deckrecord).strip() == "/":
            # For some keywords, at least PVTO, an empty record like
            # this signifies that we jump to the next table, and
            # for PVTO, this counter variable will be used as PVTNUM
            emptyrecord_counter += 1
            continue
        recdict = parse_opmio_deckrecord(deckrecord, keyword, renamer=renamer)
        if emptyrecordcountername is not None:
            recdict[emptyrecordcountername] = emptyrecord_counter
        if recordcountername is not None:
            recdict[recordcountername] = record_counter
        # Now some keywords have an arbitrary amount of data for a record, f.ex.
        # PVTO, where multiple undersaturated lines can be added. We want
        # to unroll this data. We detect this situation by the item name "DATA" in
        # the json files, and that its value is of list type:
        if "DATA" in recdict and isinstance(recdict["DATA"], list):
            assert renamer is not None
            # If DATA is sometimes used for something else in the jsons, redo this.
            data_dim = len(renamer["DATA"])  # The renamers must be in sync with json!
            data_chunks = int(len(recdict["DATA"]) / data_dim)
            try:
                data_reshaped = np.reshape(recdict["DATA"], (data_chunks, data_dim))
            except ValueError as err:
                raise ValueError(
                    (
                        f"Wrong number count for keyword {keyword}. \n"
                        "Either your keyword is wrong, or your data is wrong"
                    )
                ) from err
            data_df = pd.DataFrame(columns=renamer["DATA"], data=data_reshaped)
            # Assign the remaining items from the parsed dict to the dataframe:
            for key, value in recdict.items():
                if key != "DATA":
                    data_df[key] = value
            records.append(data_df)
            record_counter += 1
        else:
            records.append(recdict)
            record_counter += 1
    if isinstance(records[0], pd.DataFrame):
        dframe = pd.concat(records)
    else:  # records contain lists.
        dframe = pd.DataFrame(data=records)
    return dframe.reset_index(drop=True)


def parse_opmio_deckrecord(
    record: "opm.libopmcommon_python.DeckRecord",
    keyword: str,
    itemlistname: str = "items",
    recordindex: Optional[int] = None,
    renamer: Optional[Union[Dict[str, str], Dict[str, Union[str, List[str]]]]] = None,
) -> Dict[str, Any]:
    """
    Parse an opm.io.DeckRecord belonging to a certain keyword

    Args:
        record: Record be parsed
        keyword: Which Eclipse keyword this belongs to
        itemlistname: The key in the json dict that describes the items,
            typically 'items' or 'records'
        recordindex: For keywords where itemlistname is 'records', this is a
            list index to the "record".
            Beware, there are multiple concepts here for what we call a record.
        renamer: If supplied, this dictionary will be used to remap
            the keys in returned dict. For items with name DATA, the dict
            value is assumed to be a list of strings to be mapped to
            each subitem.
    """
    if keyword not in OPMKEYWORDS:
        raise ValueError(f"Keyword {keyword} not supported by common.py")

    rec_dict: Dict[str, Any] = {}

    if recordindex is None:  # Beware, 0 is different from None here.
        itemlist = OPMKEYWORDS[keyword][itemlistname]
    else:
        itemlist = OPMKEYWORDS[keyword][itemlistname][recordindex]

    # Loop over the items in the "items" section of the json keyword description.
    # Usually these items refer to one number/value in the deck record ("one line")
    # but for some keywords there are more values, like for PVTO
    for item_idx, jsonitem in enumerate(itemlist):
        item_name = jsonitem["name"]
        try:
            defaulted = record[item_idx].defaulted
        except IndexError:
            # Workaround for missing default in json for WLIST item WELLS (empty string)
            defaulted = True
        if not defaulted:
            if len(record[item_idx]) == 1:
                # The DeckItem attribute .value is only present if there is an
                # explicit statement "from opm.io.deck import DeckKeyword"
                # in this file.
                rec_dict[item_name] = record[item_idx].value
            else:
                try:
                    rec_dict[item_name] = record[item_idx].get_raw_data_list()
                except ValueError:
                    # Will get here for string lists:
                    rec_dict[item_name] = record[item_idx].get_data_list()
                # (the caller is responsible for unrolling this list with
                # correct naming of elements)

                # When we parse a list, some values in it can be defaulted, for
                # which the default values are not provided in the JSON. Return
                # these defaulted values as NaN's for the calling code to fix.
                for idx in range(len(record[item_idx])):
                    # This code is using a private attribute of an
                    # OPM DeckItem. A better solution has not yet
                    # been found in the OPM API. See also
                    # https://github.com/OPM/opm-common/issues/2598
                    if record[item_idx].__defaulted(idx):
                        rec_dict[item_name][idx] = np.nan
        else:
            rec_dict[item_name] = jsonitem.get("default", None)

    if renamer:
        renamed_dict: Dict[str, Any] = {}
        for key, value in rec_dict.items():
            if key in renamer and not isinstance(renamer[key], list):
                renamed_dict[renamer[key]] = value  # type: ignore
            else:
                renamed_dict[key] = value
        return renamed_dict
    return rec_dict


def parse_opmio_date_rec(record: "opm.io.DeckRecord") -> datetime.date:
    """Parse a opm.io.DeckRecord under a DATES or START keyword in a deck."""
    day = record[0].get_int(0)
    month = record[1].get_str(0)
    year = record[2].get_int(0)
    return datetime.date(year=year, month=parse_ecl_month(month), day=day)


def parse_opmio_tstep_rec(record: "opm.io.DeckRecord") -> List[Union[float, int]]:
    """Parse a record with TSTEP data

    Return:
        list of floats or ints
    """
    return record[0].get_raw_data_list()


def merge_zones(
    df: pd.DataFrame, zonedict: dict, zoneheader: str = "ZONE", kname: str = "K1"
) -> pd.DataFrame:
    """Merge in a column with zone names, from a dictionary mapping
    k-index to zone name. If the zonemap is not covering all
    zones, cells will be filled with NaN.

    Args:
        df: Dataframe where we should augment a column
        zonedict: Dictionary with integer keys pointing to strings
            with zone names.
        zoneheader: Name of the result column merged in,
            default: ZONE
        kname: Column header in your dataframe that maps to dictionary keys.
            default K1
    """
    assert isinstance(zonedict, dict)
    assert isinstance(zoneheader, str)
    assert isinstance(kname, str)
    assert isinstance(df, pd.DataFrame)
    if not zonedict:
        logger.warning("Can't merge in empty zone information")
        return df
    if zoneheader in df:
        logger.error(
            "Column name %s already exists, refusing to merge in any more", zoneheader
        )
        return df
    if kname not in df:
        logger.error("Can't merge on non-existing column %s", kname)
        return df
    zone_df = pd.DataFrame.from_dict(zonedict, orient="index", columns=[zoneheader])
    zone_df.index.name = "K"
    zone_df.reset_index(inplace=True)

    df[zoneheader] = df[kname].map(zonedict)
    return df


def comment_formatter(multiline: Optional[str], prefix: str = "-- ") -> str:
    """Prepends comment characters to every line in input

    If nothing is supplied, an empty string is returned.

    Args:
        multiline: String that can contain newlines
        prefix: Comment characters to prepend every line with
            Default is the Eclipse comment syntax '-- '

    Returns:
        string, with newlines preserved, and where each line
            starts with the given prefix. Always ends with a newline.
    """
    if multiline is None or not multiline.strip():
        return ""
    return (
        "\n".join([prefix + line.strip() for line in multiline.splitlines()]).strip()
        + "\n"
    )


def handle_wanted_keywords(
    wanted: Optional[List[str]],
    deck: "opm.io.Deck",
    supported: List[str],
    modulename: str = "",
) -> List[str]:
    """Handle three list of keywords, wanted, available and supported

    Args:
        keywords: None, or list of strings of user-requested keywords
        deck: Used to query which data is available
        supported: Keywords that are supported by the module
        modulename: Name of the module calling this function, used in logging
    """
    if isinstance(wanted, str):
        wanted = [wanted]
    if wanted is None or (wanted[0] is None and len(wanted) == 1):
        # By default, select all supported keywords:
        keywords = supported
    else:
        # Warn if some keywords are unsupported:
        not_supported: Set[str] = set(wanted) - set(supported)
        if not_supported:
            logger.warning(
                "Requested keyword(s) not supported by ecl2df.%s: %s",
                modulename,
                str(not_supported),
            )
        # Reduce to only supported keywords:
        keywords = list(set(wanted) - set(not_supported))
        # Warn if some requested keywords are not in deck:
        keywords_in_deck = [keyword for keyword in keywords if keyword in deck]
        not_in_deck = set(keywords) - set(keywords_in_deck)
        if not_in_deck:
            logger.warning(
                "Requested keyword(s) not present in deck: %s", str(not_in_deck)
            )
    # Reduce again to only present keywords, but without warning:
    keywords = [keyword for keyword in keywords if keyword in deck]

    return keywords


def fill_reverse_parser(
    parser: argparse.ArgumentParser, modulename: str, defaultoutputfile: str
):
    """A standardized submodule parser for the command line utility
    to produce Eclipse include files from a CSV file.

    Arguments:
        parser: parser to fill with arguments
        modulename: Will be included in the help text
        defaultoutputfile: Default output filename
    """
    parser.add_argument(
        "csvfile", help="Name of CSV file with " + modulename + " data on ecl2df format"
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help=(
            "Name of output Eclipse include file file, default "
            + defaultoutputfile
            + ". "
            "Use '-' for stdout."
        ),
        default=defaultoutputfile,
    )
    parser.add_argument(
        "-k",
        "--keywords",
        nargs="+",
        help=(
            "List of " + modulename + " keywords to include. "
            "If not supplied, all supported and found keywords will be included."
        ),
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def df2ecl(
    dataframe: pd.DataFrame,
    keywords: Optional[Union[str, List[str], List[Optional[str]]]] = None,
    comments: Optional[Dict[str, str]] = None,
    supported: Optional[List[str]] = None,
    consecutive: Optional[str] = None,
    filename: Optional[str] = None,
) -> str:
    """Generate Eclipse include strings from dataframes in ecl2df format

    Args:
        dataframe: Dataframe with Eclipse data on ecl2df format.
        keywords: List of keywords to include. Will be reduced
            to the set of keywords available in dataframe and to those supported
        comments: Dictionary indexed by keyword with comments to be
            included pr. keyword. If a key named "master" is present
            it will be used as a master comment for the outputted file.
        supported: List of strings of keywords which are
            supported in this invocation of this function.
        consecutive: Column name for which we require the
            numbers to be consecutive. Typically PVTNUM, EQLNUM, SATNUM.
        filename: If supplied, the generated text will also be dumped
            to file.

    Returns:
        string that can be used as an include file for Eclipse.
    """
    from_module = inspect.stack()[1]
    calling_module = inspect.getmodule(from_module[0])
    if dataframe.empty:
        raise ValueError("Empty dataframe")
    if consecutive is not None and consecutive in dataframe:
        if not (
            min(dataframe[consecutive]) == 1
            and len(dataframe[consecutive].unique()) == max(dataframe[consecutive])
        ):
            logger.critical(
                "%s inconsistent in input dataframe, got the values %s",
                consecutive,
                str(dataframe[consecutive].unique()),
            )
            raise ValueError

    # "KEYWORD" must always be in the dataframe:
    if "KEYWORD" not in dataframe:
        raise ValueError("KEYWORD must be in the dataframe")

    if comments is None:
        comments = {}
    if keywords is None or isinstance(keywords, str):
        keywords = [keywords]
    keywords_in_frame = set(dataframe["KEYWORD"])
    if keywords[0] is None and len(keywords) == 1:
        assert supported is not None
        keywords = supported
    else:
        # Warn if some keywords are unsupported:
        assert keywords is not None
        assert supported is not None
        not_supported: Set[Optional[str]] = set(keywords) - set(supported)
        if not_supported:
            logger.warning(
                "Requested keyword(s) not supported by %s: %s",
                calling_module.__name__,  # type: ignore
                str(not_supported),
            )
        # Warn if some requested keywords are not in frame:
        not_in_frame = set(keywords) - keywords_in_frame
        if not_in_frame:
            logger.warning(
                "Requested keyword(s) not present in dataframe: %s", str(not_in_frame)
            )
    keywords = [
        keyword
        for keyword in keywords  # user supplied list defines the print order
        if keyword in supported and keyword in keywords_in_frame
    ]
    if not keywords:
        # Nothing to do

        return ""
    string = ""
    ecl2df_header = (
        "Output file printed by "
        + calling_module.__name__  # type: ignore
        + " "
        + __version__
        + "\n"
        + " at "
        + str(datetime.datetime.now())
    )
    string += comment_formatter(ecl2df_header)
    string += "\n"
    if "master" in comments:
        string += comment_formatter(comments["master"])
    for keyword in keywords:
        # Construct the associated function names
        function_name = "df2ecl_" + keyword.lower()
        function = getattr(calling_module, function_name)
        if keyword in comments:
            string += function(dataframe, comments[keyword])
        else:
            string += function(dataframe)

    if filename is not None:
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        Path(filename).write_text(string, encoding="utf-8")
    return string


def runlength_eclcompress(string: str, sep: str = "  ") -> str:
    """Compress a string of space-separated elements so that

       2 2 2 2 2 3 3 4

    becomes

       5*2 2*3 4

    which is the format supported by Eclipse. The input
    string must be splittable with split(). Any newlines
    will be replaced by a space prior to split().

    See https:///en.wikipedia.org/wiki/Run-length_encoding

    Args:
        string (str): String of space-separated elements to compress
        sep (str): Separator string used in output. Separator string in
            input is not conserved, but it must be compatible with
            str.split()
    Returns:
        string, shorter or equal-length to the input.
    """
    compresseddata = []
    stringlist = string.replace("\n", " ").split()
    for _, group in itertools.groupby(stringlist):
        equalvalues = list(group)
        if len(equalvalues) > 1:
            compresseddata += [str(len(equalvalues)) + "*" + str(equalvalues[0])]
        else:
            compresseddata += [sep.join(equalvalues)]
    return sep.join(compresseddata)


def stack_on_colnames(
    dframe: pd.DataFrame,
    sep: str = "@",
    stackcolname: str = "DATE",
    inplace: bool = True,
) -> pd.DataFrame:
    """For a dataframe where some columns are multilevel, but where
    the second level is encoded in the column name, this function
    will stack the dataframe by putting the second level of the column
    multiindex into its own column, best understood by this example:

    A dframe like this

       ===== =============== ==============
       PORV   OWC@2000-01-01 OWC@2020-01-01
       ===== =============== ==============
       100       1000          990
       ===== =============== ==============

    will be stacked to

       ====  ====  ==========
       PORV  OWC   DATE
       ====  ====  ==========
       100   1000  2000-01-01
       100   990   2020-01-01
       ====  ====  ==========

    (for the defaults values for *sep* and *stackcolname*)

    Column order is not guaranteed

    Args:
        dframe: A dataframe to stack
        sep: The separator that is used in dframe.columns to define
            the multilevel column names.
        stackcolname: Used as column name for the second level
            of the column multiindex
    """
    if not inplace:
        dframe = dframe.copy()
    tuplecolumns = list(map(lambda x: tuple(x.split(sep)), dframe.columns))
    if max(map(len, tuplecolumns)) < 2:
        logger.info("No columns to stack")
        return dframe
    dframe.columns = pd.MultiIndex.from_tuples(
        tuplecolumns, names=["dummy", stackcolname]
    )
    dframe = dframe.stack()
    staticcols = [col[0] for col in tuplecolumns if len(col) == 1]
    dframe[staticcols] = dframe[staticcols].fillna(method="ffill")
    dframe.reset_index(inplace=True)
    # Drop rows stemming from the NaNs in the second tuple-element for
    # static columns:
    dframe.dropna(axis="index", subset=["DATE"], inplace=True)
    del dframe["level_0"]
    dframe.index.name = ""
    return dframe


def is_color(input_string: str) -> bool:
    """Checks if the input string is a valid color.
    That is six-digit hexadecimal, three-digit hexadecimal or
    given as an SVG color keyword name
    """
    if input_string.lower() in SVG_COLOR_NAMES:
        return True

    regex = "^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$"
    return bool(re.match(regex, input_string))


def parse_lyrfile(filename: str) -> Optional[List[Dict[str, Any]]]:
    """Return a list of dicts representation of the lyr file.

    The lyr file contains data of the following format,
    where the color code is optional::

      'ZoneA' 1-4     #FFE5F7
      'ZoneB' 5       red

    Description of the lyr format:
    https://resinsight.org/3d-main-window/formations/#formation-names-description-files-_lyr_

    The output has the following format::

        [
            {
                "name": "ZoneA",
                "from_layer": 1,
                "to_layer": 4,
                "color": "#FFE5F7"
            },
            {
                "name": "ZoneB",
                "span": 5,
                "color": "red"
            }
        ]

    Args:
        filename: Absolute path to a lyr file

    Returns:
        A list of dictionaries representing the information in the lyr file.

    """  # noqa

    zonelines = Path(filename).read_text(encoding="utf-8").splitlines()

    # Remove comments, support both "--" and "#":
    zonelines = [line.split("--")[0].strip() for line in zonelines]
    zonelines = [line for line in zonelines if line and not line.startswith("#")]

    lyrlist: List[Dict[str, Any]] = []
    for line in zonelines:
        try:
            linesplit = shlex.split(line)
            zonedict: Dict[str, Any] = {"name": linesplit[0]}
            zone_color = linesplit.pop(-1) if is_color(linesplit[-1]) else None
            if zone_color is not None:
                zonedict["color"] = zone_color

            numbers = " ".join(linesplit[1:]).split("-")
            if len(numbers) == 2:
                from_layer, to_layer = int(numbers[0]), int(numbers[1])
                if from_layer <= to_layer:
                    zonedict["from_layer"] = from_layer
                    zonedict["to_layer"] = to_layer
                else:
                    logger.error("From_layer higher than to_layer")
                    raise ValueError()
            elif len(numbers) == 1:
                zonedict["span"] = int(numbers[0])
            else:
                raise ValueError()
            lyrlist.append(zonedict)
        except ValueError:
            logger.error("Could not parse lyr file %s", filename)
            logger.error("Failed on content: %s", line)
            return None
    return lyrlist


def convert_lyrlist_to_zonemap(lyrlist: List[Dict[str, Any]]) -> Dict[int, str]:
    """Returns a layer to zone map as a dictionary

    Args:
        lyrlist: list of dicts coming from parse_lyrfile()
    Returns:
        Layer to zone mapping {1: "zoneA", 2: "zoneB"}
    """
    if lyrlist is None:
        return None
    zonemap = {}
    for idx, zonedict in enumerate(lyrlist):
        if "span" in zonedict:
            from_layer = lyrlist[idx - 1]["to_layer"] + 1 if idx > 0 else 1
            to_layer = from_layer + zonedict["span"]
        else:
            from_layer = zonedict["from_layer"]
            to_layer = zonedict["to_layer"]
        zonemap.update(dict.fromkeys(range(from_layer, to_layer + 1), zonedict["name"]))
    return zonemap


def get_wells_matching_template(template: str, wells: list):
    """Returns the wells in the list that is matching the template
    containing wilcard characters. The wildcard charachters supported
    are * to match zero or more charachters and ? to match a single
    non-empty character. Any wildcards in the beginning of the template
    must be preceded with a \\.

    Well name templates starting with ? might be allowed in Eclipse
    in some contexts, but not in all and will not be permitted here to
    avoid confusion.

    Args:
        template: well name template with wildcard characters
        wells: list of wells to match against the template

    Returns:
        List of matched wells
    """
    if template.startswith("*") or template.startswith("?"):
        raise ValueError(
            "Well template not allowed to start with a wildcard character: "
            f"Must be preceded with a \\: {template}"
        )
    if template.startswith("\\"):
        # Note that the two \\ are actually read as one and
        # this will return True for f.ex '\*P1'
        template = template[1:]
    regex = template.replace("*", ".*").replace("?", ".")
    return [well for well in wells if bool(re.match(regex, well))]



"""Test module for nnc2df"""

import os
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from ecl2df import ecl2csv, gruptree
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed, nothing relevant in here then",
        allow_module_level=True,
    )

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
# EIGTHCELLS is to be used in test_main_subparser when #356 is solved:
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_eightcells_dataset():
    """Test Eightcells dataset"""
    eclfiles = EclFiles(EIGHTCELLS)
    gruptree_df = gruptree.df(eclfiles.get_ecldeck())

    expected_dframe = pd.DataFrame(
        [
            ["2000-01-01", "FIELD", "GRUPTREE", np.nan],
            ["2000-01-01", "OP1", "WELSPECS", "OPS"],
            ["2000-01-01", "OPS", "GRUPTREE", "FIELD"],
        ],
        columns=["DATE", "CHILD", "KEYWORD", "PARENT"],
    )
    expected_dframe["DATE"] = pd.to_datetime(expected_dframe["DATE"])
    pd.testing.assert_frame_equal(gruptree_df, expected_dframe, check_dtype=False)


def test_gruptree2df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(REEK)
    grupdf = gruptree.df(eclfiles.get_ecldeck())

    assert not grupdf.empty
    assert len(grupdf["DATE"].unique()) == 5
    assert len(grupdf["CHILD"].unique()) == 11
    assert len(grupdf["PARENT"].dropna().unique()) == 3
    assert set(grupdf["KEYWORD"].unique()) == set(["GRUPTREE", "WELSPECS"])

    grupdfnowells = gruptree.df(eclfiles.get_ecldeck(), welspecs=False)

    assert len(grupdfnowells["KEYWORD"].unique()) == 1
    assert grupdf["PARENT"].dropna().unique()[0] == "FIELD"
    assert grupdf["KEYWORD"].unique()[0] == "GRUPTREE"


def test_str2df():
    """Test when we send in a string directly"""
    schstr = """
GRUPTREE
 'OPWEST' 'OP' /
 'OP' 'FIELD' /
 'FIELD' 'AREA' /
 'AREA' 'NORTHSEA' /
/

WELSPECS
 'OP1' 'OPWEST' 41 125 1759.74 'OIL' 0.0 'STD' 'SHUT' 'YES'  0  'SEG' /
/

"""
    deck = EclFiles.str2deck(schstr)
    grupdf = gruptree.df(deck)
    assert grupdf.dropna().empty  # the DATE is empty

    # This is only available if GRUPNET is also there
    assert "TERMINAL_PRESSURE" not in grupdf

    withstart = gruptree.df(deck, startdate="2019-01-01")
    assert not withstart.dropna().empty
    assert len(withstart) == 6


def test_grupnet_rst_docs(tmp_path):
    """Provide the input and output for the examples in the RST documentation"""
    os.chdir(tmp_path)
    schstr = """
START
 01 'JAN' 2000 /

SCHEDULE

GRUPTREE
 'OPEAST' 'OP' /
 'OPWEST' 'OP' /
 'INJEAST' 'WI' /
 'OP' 'FIELD' /
 'WI' 'FIELD' /
 'FIELD' 'AREA' /
 'AREA' 'NORTHSEA' /
/

GRUPNET
  'FIELD' 90 /
  'OPWEST' 100 /
/

WELSPECS
 'OP1'  'OPWEST'  41 125 1759.74 'OIL' 0.0 'STD' 'SHUT' 'YES'  0  'SEG' /
 'OP2'  'OPEAST'  43 122 1776.01 'OIL' 0.0 'STD' 'SHUT' 'YES'  0  'SEG' /
 'INJ1' 'INJEAST' 33 115 1960.21 'OIL' 0.0 'STD' 'SHUT' 'YES'  0  'SEG' /
/

"""
    deck = EclFiles.str2deck(schstr)
    grupdf = gruptree.df(deck)
    grupdf[["DATE", "CHILD", "PARENT", "KEYWORD"]].to_csv("gruptree.csv", index=False)
    grupdf.to_csv("gruptreenet.csv", index=False)
    grup_dict = gruptree.edge_dataframe2dict(grupdf)
    print("Copy and paste into RST files:")
    print(str(gruptree.tree_from_dict(grup_dict[0])))

    assert (
        str(gruptree.tree_from_dict(grup_dict[0])).strip()
        == """
NORTHSEA
└── AREA
    └── FIELD
        ├── OP
        │   ├── OPEAST
        │   │   └── OP2
        │   └── OPWEST
        │       └── OP1
        └── WI
            └── INJEAST
                └── INJ1
    """.strip()
    )


def test_grupnetdf():
    """Test making a dataframe from a GRUPTREE string"""
    schstr = """
GRUPTREE
 'OPWEST' 'OP' /
 'OP' 'FIELD' /
 'WI' 'FIELD' /
 'FIELD' 'AREA' /
 'AREA' 'NORTHSEA' /
/

GRUPNET
  'FIELD' 90 /
  'OPWEST' 100 /
/

"""
    deck = EclFiles.str2deck(schstr)
    grupdf = gruptree.df(deck, startdate="2000-01-01")
    print(grupdf)
    assert "TERMINAL_PRESSURE" in grupdf
    assert 90 in grupdf["TERMINAL_PRESSURE"].values
    assert 100 in grupdf["TERMINAL_PRESSURE"].values


@pytest.mark.parametrize(
    "dicttree, expected_str",
    [
        ({}, ""),
        ({"foo": {}}, "foo"),
        ({"foo": {"bar": {}}}, "foo\n└── bar"),
        ({"foo": {"bar": {}, "com": {}}}, "foo\n├── bar\n└── com"),
        # Test sorting:
        ({"foo": {"com": {}, "bar": {}}}, "foo\n├── bar\n└── com"),
        # Two levels:
        (
            {"foo": {"bar": {}, "com": {"fjooo": {}}}},
            """
foo
├── bar
└── com
    └── fjooo""",
        ),
        # Integers as node names:
        ({1: {2: {}}}, "1\n└── 2"),
        # More complex structure:
        # Note: Node names cannot be duplicated, even on
        # unique branches:
        (
            {
                "foo": {
                    "bar": {},
                    "com": {"fjooo": {}},
                    "bart": {},
                    "comt": {"fjooot": {}},
                }
            },
            """
foo
├── bar
├── bart
├── com
│   └── fjooo
└── comt
    └── fjooot
            """,
        ),
    ],
)
def test_tree_from_dict(dicttree, expected_str):
    assert str(gruptree.tree_from_dict(dicttree)).strip() == expected_str.strip()


def test_dict2treelib_deprecated():
    """dict2treelib is deprecated and replaced by tree_from_dict()"""
    with pytest.warns(FutureWarning):
        gruptree.dict2treelib("foo", {"bar": {}})


@pytest.mark.parametrize(
    "schstr, expected_dframe, expected_tree",
    [
        (
            """
GRUPTREE
 'OP' 'FIELD'/
/

GRUPNET
  'FIELD' 90 /
  'OP' 100 /
/
        """,
            pd.DataFrame(
                [
                    {"CHILD": "FIELD", "PARENT": None, "TERMINAL_PRESSURE": 90},
                    {"CHILD": "OP", "PARENT": "FIELD", "TERMINAL_PRESSURE": 100},
                ]
            ),
            """
FIELD
└── OP
""",
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """
GRUPTREE
 'OP' 'FIELDA'/
/

GRUPNET
  'FIELDA' 90 /
  'OP' 100 /
  'FIELDB' 80 /   -- This is ignored when it is not in the GRUPTREE!
/
        """,
            pd.DataFrame(
                [
                    {"CHILD": "FIELDA", "PARENT": None, "TERMINAL_PRESSURE": 90},
                    {"CHILD": "OP", "PARENT": "FIELDA", "TERMINAL_PRESSURE": 100},
                ]
            ),
            """
FIELDA
└── OP
""",
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """
GRUPTREE
 'OP' 'FIELDA'/
 'OPX' 'FIELDB' /
/

GRUPNET
  'FIELDA' 90 /
  'OP' 100 /
  'FIELDB' 80 /
/
        """,
            pd.DataFrame(
                [
                    {"CHILD": "FIELDB", "PARENT": None, "TERMINAL_PRESSURE": 80},
                    {"CHILD": "FIELDA", "PARENT": None, "TERMINAL_PRESSURE": 90},
                    {"CHILD": "OP", "PARENT": "FIELDA", "TERMINAL_PRESSURE": 100},
                    {"CHILD": "OPX", "PARENT": "FIELDB", "TERMINAL_PRESSURE": None},
                ]
            ),
            """
FIELDA
└── OP
FIELDB
└── OPX
""",
        )
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    ],
)
def test_grupnetroot(schstr, expected_dframe, expected_tree):
    """Test that terminal pressure of the tree root can be
    included in the dataframe (with an empty parent)"""
    deck = EclFiles.str2deck(schstr)
    grupdf = gruptree.df(deck, startdate="2000-01-01")
    non_default_columns = ["CHILD", "PARENT", "TERMINAL_PRESSURE"]
    pd.testing.assert_frame_equal(
        grupdf[non_default_columns]
        .sort_values(["CHILD", "PARENT"])
        .reset_index(drop=True),
        expected_dframe.sort_values(["CHILD", "PARENT"]).reset_index(drop=True),
        check_dtype=False,
    )
    treelist = gruptree.edge_dataframe2dict(grupdf)
    # Merge strings for all trees (if multiple roots)
    strtrees = [str(gruptree.tree_from_dict(tree)) for tree in treelist]
    strtrees.sort()  # Avoid flaky test due to sorting
    treelibtree = "".join(strtrees)
    assert treelibtree.strip() == expected_tree.strip()


def test_multiple_roots():
    """Test edge_dataframe2dict with multiple roots"""
    answer = [
        {"FIELDA": {"PLATA": {}}},
        {"FIELDB": {"PLATB": {}}},
    ]
    edges = pd.DataFrame(
        [
            {"CHILD": "FIELDA", "PARENT": None},
            {"CHILD": "FIELDB", "PARENT": None},
            {"CHILD": "PLATA", "PARENT": "FIELDA"},
            {"CHILD": "PLATB", "PARENT": "FIELDB"},
        ]
    )
    assert gruptree.edge_dataframe2dict(edges) == answer

    # Same result if the dummy rows for the roots are omitted:
    edges_noroots = pd.DataFrame(
        [
            {"CHILD": "PLATA", "PARENT": "FIELDA"},
            {"CHILD": "PLATB", "PARENT": "FIELDB"},
        ]
    )
    assert gruptree.edge_dataframe2dict(edges_noroots) == answer

    # And order does not matter, should be sorted on root node label:
    edges_noroots = pd.DataFrame(
        [
            {"CHILD": "PLATB", "PARENT": "FIELDB"},
            {"CHILD": "PLATA", "PARENT": "FIELDA"},
        ]
    )
    assert gruptree.edge_dataframe2dict(edges_noroots) == answer

    # The function tree_from_dict should be called with one tree at a time:
    with pytest.raises(ValueError, match="single tree"):
        gruptree.tree_from_dict({"foo": 1, "bar": 2})


@pytest.mark.parametrize(
    "dframe, expected",
    [
        (pd.DataFrame(), [{}]),
        pytest.param(
            pd.DataFrame([{"FOO": "BAR"}]),
            [{}],
            marks=pytest.mark.xfail(raises=KeyError),
            # "PARENT" is required.
        ),
        pytest.param(
            pd.DataFrame([{"PARENT": "A"}]),
            [{}],
            marks=pytest.mark.xfail(raises=KeyError)
            # "CHILD" is also required
        ),
        (pd.DataFrame([{"PARENT": "A", "CHILD": "B"}]), [{"A": {"B": {}}}]),
        (
            pd.DataFrame(
                [{"PARENT": "A", "CHILD": "B"}, {"PARENT": "B", "CHILD": "C"}]
            ),
            [{"A": {"B": {"C": {}}}}],
        ),
        (
            # DATE is not used
            pd.DataFrame([{"PARENT": "A", "CHILD": "B", "DATE": "2000-01-01"}]),
            [{"A": {"B": {}}}],
        ),
        pytest.param(
            # DATE is not used, but will be checked that it is the same:
            pd.DataFrame(
                [
                    {"PARENT": "A", "CHILD": "B", "DATE": "2000-01-01"},
                    {"PARENT": "A", "CHILD": "B", "DATE": "2040-01-01"},
                ]
            ),
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_edge_dataframe2dict(dframe, expected):
    assert gruptree.edge_dataframe2dict(dframe) == expected


def test_emptytree_strdeck():
    """Test empty schedule sections. Don't want to crash"""
    schstr = ""
    deck = EclFiles.str2deck(schstr)
    grupdf = gruptree.df(deck)
    assert grupdf.empty
    gruptreedict = gruptree.edge_dataframe2dict(grupdf)
    assert not gruptreedict[0]

    treelibtree = gruptree.tree_from_dict(gruptreedict[0])
    # Returning an empty string and not a treelib.Tree() is
    # a workaround for a limitation in treelib.
    assert treelibtree == ""


def test_emptytree_commandlinetool(tmp_path, mocker, caplog):
    os.chdir(tmp_path)
    Path("EMPTY.DATA").write_text("")
    mocker.patch("sys.argv", ["ecl2csv", "gruptree", "--prettyprint", "EMPTY.DATA"])
    ecl2csv.main()
    assert "No tree data to prettyprint" in caplog.text


def test_cli_nothing_to_do(mocker, capsys):
    """Test that the client says nothing to do when DATA is supplied, but no action."""
    mocker.patch("sys.argv", ["ecl2csv", "gruptree", "EMPTY.DATA"])
    with pytest.raises(SystemExit):
        ecl2csv.main()
    assert "Nothing to do" in capsys.readouterr().out


def test_tstep():
    """Test that we can parse a deck using TSTEP for timestepping"""
    schstr = """
GRUPTREE
 'OPWEST' 'OP' /
 'OP' 'FIELD' /
 'FIELD' 'AREA' /
 'AREA' 'NORTHSEA' /
/

TSTEP
  1 /

WELSPECS
 'OP1' 'OPWEST' 41 125 1759.74 'OIL' 0.0 'STD' 'SHUT' 'YES'  0  'SEG' /
/

"""
    deck = EclFiles.str2deck(schstr)
    grupdf = gruptree.df(deck)
    assert len(grupdf["DATE"].unique()) == 2
    print(grupdf)


def test_main(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "gruptree.csv"
    mocker.patch("sys.argv", ["ecl2csv", "gruptree", REEK, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty


def test_prettyprint_commandline(mocker, capsys):
    """Test pretty printing via command line interface"""
    mocker.patch("sys.argv", ["ecl2csv", "gruptree", REEK, "--prettyprint"])
    ecl2csv.main()
    stdout = capsys.readouterr().out.strip()
    print(stdout)
    assert (
        """
Date: 2000-01-01
GRUPTREE trees:
FIELD
├── OP
└── WI


Date: 2000-02-01
GRUPTREE trees:
FIELD
├── OP
│   ├── OP_1
│   ├── OP_2
│   └── OP_3
└── WI
    └── WI_1


Date: 2000-06-01
GRUPTREE trees:
FIELD
├── OP
│   ├── OP_1
│   ├── OP_2
│   └── OP_3
└── WI
    ├── WI_1
    └── WI_2


Date: 2001-01-01
GRUPTREE trees:
FIELD
├── OP
│   ├── OP_1
│   ├── OP_2
│   ├── OP_3
│   ├── OP_4
│   └── OP_5
└── WI
    ├── WI_1
    └── WI_2


Date: 2001-03-01
GRUPTREE trees:
FIELD
├── OP
│   ├── OP_1
│   ├── OP_2
│   ├── OP_3
│   ├── OP_4
│   └── OP_5
└── WI
    ├── WI_1
    ├── WI_2
    └── WI_3
""".strip()
        in stdout
    )


def test_main_subparser(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "gruptree.csv"
    mocker.patch("sys.argv", ["ecl2csv", "gruptree", "-v", REEK, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty


@pytest.mark.parametrize(
    "schstr, expected_dframe, check_columns",
    [
        (
            # Changing BRANPROP
            """
DATES
  1 JAN 2000 /
/
GRUPTREE
 'TMPL_A' 'FIELD'/
/
BRANPROP
  'NODE_A'  'FIELD'  1 /
  'TMPL_A'  'NODE_A' 2 /
/
NODEPROP
  'FIELD'  20 /
  'TMPL_A'  2*  YES /
/
DATES
  1 FEB 2000 /
/
BRANPROP
  'NODE_B'  'FIELD'  3 /
  'TMPL_A'  'NODE_B' 4 /
/
        """,
            pd.DataFrame(
                [
                    ["2000-01-01", "FIELD", "GRUPTREE", np.nan, np.nan, np.nan],
                    ["2000-01-01", "TMPL_A", "GRUPTREE", "FIELD", np.nan, np.nan],
                    ["2000-01-01", "FIELD", "BRANPROP", np.nan, np.nan, 20],
                    ["2000-01-01", "NODE_A", "BRANPROP", "FIELD", 1, np.nan],
                    ["2000-01-01", "TMPL_A", "BRANPROP", "NODE_A", 2, np.nan],
                    ["2000-02-01", "FIELD", "BRANPROP", np.nan, np.nan, 20],
                    ["2000-02-01", "NODE_A", "BRANPROP", "FIELD", 1, np.nan],
                    ["2000-02-01", "NODE_B", "BRANPROP", "FIELD", 3, np.nan],
                    ["2000-02-01", "TMPL_A", "BRANPROP", "NODE_B", 4, np.nan],
                ],
                columns=[
                    "DATE",
                    "CHILD",
                    "KEYWORD",
                    "PARENT",
                    "VFP_TABLE",
                    "TERMINAL_PRESSURE",
                ],
            ),
            ["DATE", "CHILD", "KEYWORD", "PARENT", "VFP_TABLE", "TERMINAL_PRESSURE"],
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            # Changing NODEPROP
            """
DATES
  1 JAN 2000 /
/
GRUPTREE
 'TMPL_A' 'FIELD'/
/
BRANPROP
  'NODE_A'  'FIELD'  /
  'TMPL_A'  'NODE_A'  /
/
NODEPROP
  'FIELD'  20 /
  'TMPL_A'  2*  YES /
/
DATES
  1 FEB 2000 /
/
NODEPROP
  'FIELD'  22  1* YES /
/
        """,
            pd.DataFrame(
                [
                    ["2000-01-01", "FIELD", "GRUPTREE", np.nan, np.nan, np.nan],
                    ["2000-01-01", "TMPL_A", "GRUPTREE", "FIELD", np.nan, np.nan],
                    ["2000-01-01", "FIELD", "BRANPROP", np.nan, 20, "NO"],
                    ["2000-01-01", "NODE_A", "BRANPROP", "FIELD", np.nan, np.nan],
                    ["2000-01-01", "TMPL_A", "BRANPROP", "NODE_A", np.nan, "YES"],
                    ["2000-02-01", "FIELD", "BRANPROP", np.nan, 22, "YES"],
                    ["2000-02-01", "NODE_A", "BRANPROP", "FIELD", np.nan, np.nan],
                    ["2000-02-01", "TMPL_A", "BRANPROP", "NODE_A", np.nan, "YES"],
                ],
                columns=[
                    "DATE",
                    "CHILD",
                    "KEYWORD",
                    "PARENT",
                    "TERMINAL_PRESSURE",
                    "ADD_GAS_LIFT_GAS",
                ],
            ),
            [
                "DATE",
                "CHILD",
                "KEYWORD",
                "PARENT",
                "TERMINAL_PRESSURE",
                "ADD_GAS_LIFT_GAS",
            ],
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            # WELSPECS
            """
DATES
  1 JAN 2000 /
/
GRUPTREE
 'TMPL_A' 'FIELD'/
/
BRANPROP
  'NODE_A'  'FIELD'  /
  'TMPL_A'  'NODE_A'  /
/
NODEPROP
  'FIELD'  20 /
  'TMPL_A'  2*  YES /
/
WELSPECS
  'WELL_1'  'TMPL_A' 1 1 1 OIL /
  'WELL_2'  'TMPL_B' 1 1 1 OIL /
/
DATES
  1 FEB 2000 /
/
NODEPROP
  'FIELD' 22 /
/
        """,
            # TMPL_B is not in any trees. The WELSPECS line is added
            # only when there is a new GRUPTREE tree and a GRUPTREE
            # entry is added connecting TMPL_B to the FIELD node.
            # TMPL_A is in both trees, but is not repeated at the date
            # where there are two trees
            pd.DataFrame(
                [
                    ["2000-01-01", "FIELD", "GRUPTREE", np.nan],
                    ["2000-01-01", "TMPL_A", "GRUPTREE", "FIELD"],
                    ["2000-01-01", "WELL_2", "WELSPECS", "TMPL_B"],
                    ["2000-01-01", "TMPL_B", "GRUPTREE", "FIELD"],
                    ["2000-01-01", "FIELD", "BRANPROP", np.nan],
                    ["2000-01-01", "NODE_A", "BRANPROP", "FIELD"],
                    ["2000-01-01", "TMPL_A", "BRANPROP", "NODE_A"],
                    ["2000-01-01", "WELL_1", "WELSPECS", "TMPL_A"],
                    ["2000-02-01", "FIELD", "BRANPROP", np.nan],
                    ["2000-02-01", "NODE_A", "BRANPROP", "FIELD"],
                    ["2000-02-01", "TMPL_A", "BRANPROP", "NODE_A"],
                    ["2000-02-01", "WELL_1", "WELSPECS", "TMPL_A"],
                ],
                columns=["DATE", "CHILD", "KEYWORD", "PARENT"],
            ),
            ["DATE", "CHILD", "KEYWORD", "PARENT"],
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    ],
)
def test_branprop_nodeprop(schstr, expected_dframe, check_columns):
    """Testing that the gruptree dataframe works correctly
    when the schedule string contains BRANPROP and NODEPROP
    """
    deck = EclFiles.str2deck(schstr)
    dframe = gruptree.df(deck).reset_index()
    expected_dframe.DATE = pd.to_datetime(expected_dframe.DATE)
    pd.testing.assert_frame_equal(
        dframe[check_columns],
        expected_dframe[check_columns],
        check_dtype=False,
    )


def test_prettyprint(tmp_path, mocker, caplog):
    """ "Test prettyprinting with multiple dates and both
    GRUPTREE and BRANPROP trees"""
    schstr = """
DATES
  1 JAN 2000 /
/
GRUPTREE
 'TMPL_A' 'FIELD'/
/
BRANPROP
  'NODE_A'  'FIELD'  /
  'TMPL_A'  'NODE_A'  /
/
NODEPROP
  'FIELD'  20 /
  'TMPL_A'  2*  YES /
/
WELSPECS
  'WELL_1'  'TMPL_A' 1 1 1 OIL /
  'WELL_2'  'TMPL_B' 1 1 1 OIL /
/
DATES
  1 FEB 2000 /
/
NODEPROP
  'FIELD' 22 /
/
    """

    expected_prettyprint = """
Date: 2000-01-01
GRUPTREE trees:
FIELD
├── TMPL_A
│   └── WELL_1
└── TMPL_B
    └── WELL_2

BRANPROP trees:
FIELD
└── NODE_A
    └── TMPL_A
        └── WELL_1


Date: 2000-02-01
BRANPROP trees:
FIELD
└── NODE_A
    └── TMPL_A
        └── WELL_1


    """
    dframe = gruptree.df(EclFiles.str2deck(schstr))
    assert gruptree.prettyprint(dframe).strip() == expected_prettyprint.strip()



from pathlib import Path

from setuptools import find_packages, setup

try:
    from sphinx.setup_command import BuildDoc

    cmdclass = {"build_sphinx": BuildDoc}
except ImportError:
    # Skip cmdclass when sphinx is not installed (yet)

    cmdclass = {}

# Read the contents of README.md, for PyPI
LONG_DESCRIPTION = (Path(__file__).parent / "README.md").read_text()

SETUP_REQUIREMENTS = ["setuptools>=28", "setuptools_scm < 6.1"]
REQUIREMENTS = [
    "ecl",
    "numpy",
    "opm>=2020.10.2",  # NB: Pypi versions.
    "pandas<1.4.0",
    "pyarrow",
    "pyyaml>=5.1",
    "treelib",
]

TEST_REQUIREMENTS = Path("test_requirements.txt").read_text().splitlines()

DOCS_REQUIREMENTS = [
    "autoapi",
    "ipython",
    "rstcheck",
    "sphinx",
    "sphinx-argparse",
    "sphinx-autodoc-typehints",
    "sphinx_rtd_theme",
]
EXTRAS_REQUIRE = {"tests": TEST_REQUIREMENTS, "docs": DOCS_REQUIREMENTS, "ert": ["ert"]}

setup(
    name="ecl2df",
    use_scm_version={"write_to": "ecl2df/version.py"},
    cmdclass=cmdclass,
    description="Convert Eclipse 100 input and output to DataFrames",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="http://github.com/equinor/ecl2df",
    author="Håvard Berland",
    author_email="havb@equinor.com",
    license="GPLv3",
    packages=find_packages(include=["ecl2df*"]),
    package_dir={"ecl2df": "ecl2df"},
    package_data={
        "ecl2df": [
            "opmkeywords/*",
            "config_jobs/*",
            "py.typed",
            "svg_color_keyword_names.txt",
        ]
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "csv2ecl=ecl2df.csv2ecl:main",
            "ecl2csv=ecl2df.ecl2csv:main",
            "ecl2arrow=ecl2df.ecl2csv:main",
        ],
        "ert": ["ecl2df_jobs = ecl2df.hook_implementations.jobs"],
    },
    test_suite="tests",
    install_requires=REQUIREMENTS,
    setup_requires=SETUP_REQUIREMENTS,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.6",
)



"""Test module for wcon"""

import io
import subprocess
from pathlib import Path

import pandas as pd
import pytest

from ecl2df import ecl2csv, wcon
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed",
        allow_module_level=True,
    )
TESTDIR = Path(__file__).absolute().parent
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_wcon2df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(EIGHTCELLS)
    wcondf = wcon.df(eclfiles.get_ecldeck())

    assert not wcondf.empty
    assert "DATE" in wcondf  # for all data
    assert "KEYWORD" in wcondf
    for col in wcondf.columns:
        assert col == col.upper()


def test_wconhist():
    """Test WCONHIST parsing and column names"""
    wconstr = """
WCONHIST
  'FOO' 0 1 /
 /
"""
    deck = EclFiles.str2deck(wconstr)
    wconhist_df = wcon.df(deck)
    pd.testing.assert_frame_equal(
        wconhist_df,
        pd.DataFrame(
            [
                {
                    "WELL": "FOO",
                    "STATUS": "0",
                    "CMODE": "1",
                    "ORAT": 0,
                    "WRAT": 0,
                    "GRAT": 0,
                    "VFP_TABLE": 0,
                    "ALQ": 0,
                    "THP": 0,
                    "BHP": 0,
                    "NGLRAT": 0,
                    "DATE": None,
                    "KEYWORD": "WCONHIST",
                }
            ]
        ),
    )


def test_wconinjh():
    """Test WCONINJH parsing and column names"""
    wconstr = """
WCONINJH
  'FOO' 0 1 /
 /
"""
    deck = EclFiles.str2deck(wconstr)
    wconinjh_df = wcon.df(deck)
    pd.testing.assert_frame_equal(
        wconinjh_df,
        pd.DataFrame(
            [
                {
                    "WELL": "FOO",
                    "TYPE": "0",
                    "STATUS": "1",
                    "RATE": None,
                    "BHP": None,
                    "THP": None,
                    "VFP_TABLE": 0,
                    "VAPOIL_C": 0,
                    "SURFACE_OIL_FRACTION": 0,
                    "SURFACE_WATER_FRACTION": 0,
                    "SURFACE_GAS_FRACTION": 0,
                    "CMODE": "RATE",
                    "DATE": None,
                    "KEYWORD": "WCONINJH",
                }
            ]
        ),
    )


def test_wconinje():
    """Test WCONINJE parsing and column names"""
    wconstr = """
WCONINJE
  'FOO' 0 1 /
 /
"""
    deck = EclFiles.str2deck(wconstr)
    wconinje_df = wcon.df(deck)
    pd.testing.assert_frame_equal(
        wconinje_df,
        pd.DataFrame(
            [
                {
                    "WELL": "FOO",
                    "TYPE": "0",
                    "STATUS": "1",
                    "CMODE": None,
                    "RATE": None,
                    "RESV": None,
                    "BHP": 6895,
                    "THP": None,
                    "VFP_TABLE": 0,
                    "VAPOIL_C": 0,
                    "GAS_STEAM_RATIO": 0,
                    "SURFACE_OIL_FRACTION": 0,
                    "SURFACE_WATER_FRACTION": 0,
                    "SURFACE_GAS_FRACTION": 0,
                    "OIL_STEAM_RATIO": 0,
                    "DATE": None,
                    "KEYWORD": "WCONINJE",
                }
            ]
        ),
    )


def test_wconprod():
    """Test WCONPROD parsing and column names"""
    wconstr = """
WCONPROD
  'FOO' 0 1 /
 /
"""
    deck = EclFiles.str2deck(wconstr)
    wconprod_df = wcon.df(deck)
    pd.testing.assert_frame_equal(
        wconprod_df,
        pd.DataFrame(
            [
                {
                    "WELL": "FOO",
                    "STATUS": "0",
                    "CMODE": "1",
                    "ORAT": 0,
                    "WRAT": 0,
                    "GRAT": 0,
                    "LRAT": 0,
                    "RESV": 0,
                    "BHP": 1.01325,
                    "THP": 0,
                    "VFP_TABLE": 0,
                    "ALQ": 0,
                    # These E300 columns should not
                    # be regarded critical for API.
                    "E300_ITEM13": None,
                    "E300_ITEM14": None,
                    "E300_ITEM15": None,
                    "E300_ITEM16": None,
                    "E300_ITEM17": None,
                    "E300_ITEM18": None,
                    "E300_ITEM19": None,
                    "E300_ITEM20": None,
                    "DATE": None,
                    "KEYWORD": "WCONPROD",
                }
            ]
        ),
    )


def test_tstep():
    """Test that we support the TSTEP keyword"""
    schstr = """
DATES
   1 MAY 2001 /
/

WCONHIST
 'OP1' 1000  /
/

TSTEP
  1 /

WCONHIST
 'OP1' 2000 /
/

TSTEP
  2 3 /

WCONHIST
  'OP1' 3000 /
/
"""
    deck = EclFiles.str2deck(schstr)
    wcondf = wcon.df(deck)
    dates = [str(x) for x in wcondf["DATE"].unique()]
    assert len(dates) == 3
    assert "2001-05-01" in dates
    assert "2001-05-02" in dates
    assert "2001-05-07" in dates


def test_main_subparsers(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / ".TMP-wcondf.csv"
    mocker.patch("sys.argv", ["ecl2csv", "wcon", EIGHTCELLS, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty


def test_magic_stdout():
    """Test that we can pipe the output into a dataframe"""
    result = subprocess.run(
        ["ecl2csv", "wcon", "-v", "-o", "-", EIGHTCELLS],
        check=True,
        stdout=subprocess.PIPE,
    )
    df_stdout = pd.read_csv(io.StringIO(result.stdout.decode()))
    assert not df_stdout.empty



"""Provide a two-way Pandas DataFrame interface to Eclipse summary data (UNSMRY)"""
import argparse
import ctypes

# The name 'datetime' is in use by a function argument:
import datetime as dt
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import dateutil
import numpy as np
import pandas as pd
import pyarrow
import pyarrow.feather
from ecl.summary import EclSum, EclSumKeyWordVector

from ecl2df import getLogger_ecl2csv

from . import parameters
from .common import write_dframe_stdout_file
from .eclfiles import EclFiles

logger: logging.Logger = logging.getLogger(__name__)

# Frequency mnemonics for the API consumer to use:
FREQ_RAW: str = "raw"
FREQ_FIRST: str = "first"
FREQ_LAST: str = "last"
PD_FREQ_MNEMONICS: Dict[str, str] = {
    "daily": "D",
    "weekly": "W-MON",
    "monthly": "MS",
    "yearly": "YS",
    # Any frequency mnemonics not mentioned here will be
    # passed on to Pandas.
}
"""Mapping from ecl2df custom offset strings to Pandas DateOffset strings.
See
https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
"""  # noqa


def date_range(start_date: dt.date, end_date: dt.date, freq: str) -> List[dt.datetime]:
    """Wrapper for pandas.date_range to allow for extra ecl2df specific mnemonics
    'yearly', 'daily', 'weekly', mapped over to pandas DateOffsets.

    Args:
        start_date (datetime.date):
        end_date (datetime.date):
        freq (str): monthly, daily, weekly, yearly, or a Pandas date offset
            frequency.

    Returns:
        list of datetimes
    """
    try:
        return pd.date_range(
            start_date, end_date, freq=PD_FREQ_MNEMONICS.get(freq, freq)
        )
    except pd.errors.OutOfBoundsDatetime:
        return _fallback_date_range(start_date, end_date, freq)


def _ensure_date_or_none(some_date: Optional[Union[str, dt.date]]) -> Optional[dt.date]:
    """Ensures an object is either a date or None

    Args:
        some_date: string or a datetime.date

    Returns:
        None if input is None.
    """
    if some_date is None:
        return None
    if isinstance(some_date, dt.date):
        return some_date
    if some_date == "":
        return None
    if isinstance(some_date, str):
        return dateutil.parser.parse(some_date).date()  # type: ignore
    raise TypeError(f"some_date must be a string or a date, got {some_date}")


def _crop_datelist(
    eclsumsdates: List[dt.datetime],
    freq: Union[dt.date, dt.datetime, str],
    start_date: Optional[dt.date] = None,
    end_date: Optional[dt.date] = None,
) -> Union[List[dt.date], List[dt.datetime]]:
    """Helper function for resample_smry_dates, taking care of
    the special cases where the list of dates should not be resampled, but
    only cropped or returned as is.

    Args:
        eclsumsdates: list of datetimes, typically coming from EclSum.dates
        freq: Either a date or datetime, or a frequency string
            "raw", "first" or "last".
        start_date: Dates prior to this date will be cropped.
        end_date: Dates after this date will be cropped.

    Returns:
        list of datetimes.
    """
    datetimes: Union[List[dt.date], List[dt.datetime]] = []  # type: ignore
    if freq == FREQ_RAW:
        datetimes = eclsumsdates
        datetimes.sort()
        if start_date:
            # Convert to datetime (at 00:00:00)
            start_date = dt.datetime.combine(start_date, dt.datetime.min.time())
            datetimes = [x for x in datetimes if x > start_date]
            datetimes = [start_date] + datetimes
        if end_date:
            end_date = dt.datetime.combine(end_date, dt.datetime.min.time())
            datetimes = [x for x in datetimes if x < end_date]
            datetimes = datetimes + [end_date]
    elif freq == FREQ_FIRST:
        datetimes = [min(eclsumsdates).date()]
    elif freq == FREQ_LAST:
        datetimes = [max(eclsumsdates).date()]
    elif isinstance(freq, (dt.date, dt.datetime)):
        datetimes = [freq]
    return datetimes


def _fallback_date_roll(rollme: dt.datetime, direction: str, freq: str) -> dt.datetime:
    """Fallback function for rolling dates forward or backward onto a
    date frequency boundary.

    This function reimplements pandas' DateOffset.roll_forward() and backward()
    only for monthly and yearly frequency. This is necessary as Pandas does not
    support datetimes beyond year 2262 due to all datetimes in Pandas being
    represented by nanosecond accuracy.

    This function is a fallback only, to keep support for using all Pandas timeoffsets
    in situations where years beyond 2262 is not a issue."""
    if direction not in ["back", "forward"]:
        raise ValueError(f"Unknown direction {direction}")

    if freq == "yearly":
        if direction == "forward":
            if rollme <= dt.datetime(year=rollme.year, month=1, day=1):
                return dt.datetime(year=rollme.year, month=1, day=1)
            return dt.datetime(year=rollme.year + 1, month=1, day=1)
        return dt.datetime(year=rollme.year, month=1, day=1)

    if freq == "monthly":
        if direction == "forward":
            if rollme <= dt.datetime(year=rollme.year, month=rollme.month, day=1):
                return dt.datetime(year=rollme.year, month=rollme.month, day=1)
            return dt.datetime(
                year=rollme.year, month=rollme.month, day=1
            ) + dateutil.relativedelta.relativedelta(  # type: ignore
                months=1
            )
        return dt.datetime(year=rollme.year, month=rollme.month, day=1)

    raise ValueError(
        "Only yearly or monthly frequencies are "
        "supported for simulations beyond year 2262"
    )


def _fallback_date_range(start: dt.date, end: dt.date, freq: str) -> List[dt.datetime]:
    """Fallback routine for generating date ranges beyond Pandas datetime64[ns]
    year-2262 limit.

    Assumes that the start and end times already fall on a frequency boundary.
    """
    if start == end:
        return [dt.datetime.combine(start, dt.datetime.min.time())]
    if end < start:
        return []
    if freq == "yearly":
        dates = [dt.datetime.combine(start, dt.datetime.min.time())] + [
            dt.datetime(year=year, month=1, day=1)
            for year in range(start.year + 1, end.year + 1)
        ]
        if dt.datetime.combine(end, dt.datetime.min.time()) != dates[-1]:
            dates = dates + [dt.datetime.combine(end, dt.datetime.min.time())]
        return dates
    if freq == "monthly":
        dates = []
        date = dt.datetime.combine(start, dt.datetime.min.time())
        enddatetime = dt.datetime.combine(end, dt.datetime.min.time())
        while date <= enddatetime:
            dates.append(date)
            date = date + dateutil.relativedelta.relativedelta(months=1)  # type: ignore
        return dates
    raise ValueError("Unsupported frequency for datetimes beyond year 2262")


def resample_smry_dates(
    eclsumsdates: List[dt.datetime],
    freq: str = FREQ_RAW,
    normalize: bool = True,
    start_date: Optional[Union[str, dt.date]] = None,
    end_date: Optional[Union[str, dt.date]] = None,
) -> Union[List[dt.date], List[dt.datetime]]:
    """
    Resample (optionally) a list of date(time)s to a new datelist according to options.

    Based on the dates as input, a new list at a finer or coarser time density
    can be returned, on the same date range. Incoming dates can also be cropped.

    Args:
        eclsumsdates: list of datetimes, typically coming from EclSum.dates
        freq: string denoting requested frequency for
            the returned list of datetime. 'raw' will
            return the input datetimes (no resampling).
            Options for timeresampling are
            'daily', 'weekly', 'monthly' and 'yearly'. 'first' will give
            the first date (minimum),  'last' will give out the last
            date (maximum), as a list with one element. Can also be a single date.
        normalize: Whether to normalize backwards at the start
            and forwards at the end to ensure the raw
            date range is covered when resampling time. This is
            ignored if start_date or end_date is explicitly supplied.
        start_date: str or date with first date to include
            Dates prior to this date will be dropped, supplied
            start_date will always be included. Overrides
            normalized dates.
        end_date: str or date with last date to be included.
            Dates past this date will be dropped, supplied
            end_date will always be included. Overrides
            normalized dates. Overridden if freq is 'last'.
    """
    start_date = _ensure_date_or_none(start_date)
    end_date = _ensure_date_or_none(end_date)

    if freq in [FREQ_RAW, FREQ_FIRST, FREQ_LAST] or isinstance(
        freq, (dt.date, dt.datetime)
    ):
        return _crop_datelist(eclsumsdates, freq, start_date, end_date)

    # In case freq is an ISO-date(time)-string, interpret as such:
    try:
        parseddate = dateutil.parser.isoparse(freq)  # type: ignore
        return [parseddate]
    except ValueError:
        # freq is a frequency string or datetime.date (or similar)
        pass

    # These are datetime.datetime, not datetime.date
    start_smry = min(eclsumsdates)
    end_smry = max(eclsumsdates)

    # Normalize start and end date according to frequency by extending the time range.
    # [1997-11-05, 2020-03-02] and monthly frequecy
    # will be mapped to [1997-11-01, 2020-04-01]
    # For yearly frequency it will return [1997-01-01, 2021-01-01].
    offset = pd.tseries.frequencies.to_offset(PD_FREQ_MNEMONICS.get(freq, freq))
    try:
        start_normalized = offset.rollback(start_smry.date()).date()
    except pd.errors.OutOfBoundsDatetime:
        # Pandas only supports datetime up to year 2262
        start_normalized = _fallback_date_roll(start_smry, "back", freq).date()
    try:
        end_normalized = offset.rollforward(end_smry.date()).date()
    except pd.errors.OutOfBoundsDatetime:
        # Pandas only supports datetime up to year 2262
        end_normalized = _fallback_date_roll(end_smry, "forward", freq).date()

    if start_date is None:
        if normalize:
            start_date_range = start_normalized
        else:
            start_date_range = start_smry.date()
    else:
        # Normalization is not applied for explicit date
        start_date_range = start_date

    if end_date is None:
        if normalize:
            end_date_range = end_normalized
        else:
            end_date_range = end_smry.date()
    else:
        # Normalization is not applied for explicit date
        end_date_range = end_date

    datetimes = date_range(start_date_range, end_date_range, freq)

    # Convert from numpys datetime64 to datetime.date:
    dates = [x.date() for x in datetimes]

    # pd.date_range will not include random dates that do not
    # fit on frequency boundary. Force include these if
    # supplied as user arguments.
    if start_date and start_date not in dates:
        dates = [start_date] + dates
    if end_date and end_date not in dates:
        dates = dates + [end_date]
    return dates


def df(
    eclfiles: EclFiles,
    time_index: Optional[str] = None,
    column_keys: Union[List[str], str] = None,
    start_date: Optional[Union[str, dt.date]] = None,
    end_date: Optional[Union[str, dt.date]] = None,
    include_restart: bool = True,
    params: bool = False,
    paramfile: Optional[str] = None,
    datetime: bool = False,  # A very poor choice of argument name [pylint]
):
    # pylint: disable=too-many-arguments
    """
    Extract data from UNSMRY as Pandas dataframes.

    This is a thin wrapper for EclSum.pandas_frame, by adding
    support for string mnenomics for the time index.

    The dataframe is always indexed by DATE, and the datatype for the
    index will usually be datetime64[ns] as long as all dates are
    before year 2262. If a longer time range is detected, the index.dtype
    will be object, and consisting of datetime.datetime() objects. The index
    is always named "DATE".

    Arguments:
        eclfiles: EclFiles object representing the Eclipse deck. Alternatively
           an EclSum object.
        time_index: string indicating a resampling frequency,
           'yearly', 'monthly', 'daily', 'last' or 'raw', the latter will
           return the simulated report steps (also default).
           If a list of DateTime is supplied, data will be resampled
           to these.
        column_keys: list of column key wildcards. None means everything.
        start_date: str or date with first date to include.
            Dates prior to this date will be dropped, supplied
            start_date will always be included.
        end_date: str or date with last date to be included.
            Dates past this date will be dropped, supplied
            end_date will always be included. Overriden if time_index
            is 'last'.
        include_restart: boolean sent to libecl for wheter restarts
            files should be traversed
        params: If set, parameters.txt will be attempted loaded
            and merged with the summary data.
        paramsfile: Explicit path to parameters file if autodiscovery is
            not wanted. Implies params=True
        datetime: If True, the time index of the returned DataFrame
            is always of datetime type. If not, it will be datetime
            if raw dates are requested (which are at second accuracy),
            or it will be strings in case of yearly, monthly or daily
            time frequency.

    Returns empty dataframe if there is no summary file, or if the
    column_keys are not existing.
    """
    if isinstance(column_keys, str):
        column_keys = [column_keys]

    if isinstance(eclfiles, EclSum):
        eclsum = eclfiles
    else:
        try:
            eclsum = eclfiles.get_eclsum(include_restart=include_restart)
        except OSError:
            logger.warning("Error reading summary instance, returning empty dataframe")
            return pd.DataFrame()

    time_index_arg: Optional[Union[List[dt.date], List[dt.datetime]]]
    if isinstance(time_index, str) and time_index == "raw":
        time_index_arg = resample_smry_dates(
            eclsum.dates,
            "raw",
            False,
            start_date,
            end_date,
        )
    elif isinstance(time_index, str):
        time_index_arg = resample_smry_dates(
            eclsum.dates,
            time_index,
            True,
            start_date,
            end_date,
        )
    else:
        time_index_arg = time_index

    if time_index_arg is None:
        time_index_str = ""
    if isinstance(time_index_arg, (list, np.ndarray)):
        if len(time_index_arg) < 6:
            time_index_str = str(time_index_arg)
        else:
            time_index_str = f"{time_index_arg[0:3]} … {time_index_arg[-3:]}"

    if not column_keys or not column_keys[0]:
        column_keys_str = "*"
        # column_keys = [column_keys_str]
    else:
        column_keys_str = ",".join(filter(None, column_keys))
    logger.info(
        "Requesting columns_keys: %s at time_index: %s",
        column_keys_str,
        time_index_str or "raw",
    )

    # dframe = eclsum.pandas_frame(time_index_arg, column_keys)
    dframe = _libecl_eclsum_pandas_frame(eclsum, time_index_arg, column_keys)

    logger.info(
        "Dataframe with smry data ready, %d columns and %d rows",
        len(dframe.columns),
        len(dframe),
    )
    dframe.index.name = "DATE"
    if params or paramfile:
        dframe = _merge_params(dframe, paramfile, eclfiles)

    # Add metadata as an attribute the dataframe, using experimental Pandas features:
    meta = smry_meta(eclsum)
    # Slice meta to dataframe columns:
    dframe.attrs["meta"] = {
        column_key: meta[column_key] for column_key in dframe if column_key in meta
    }

    if datetime is True:
        if dframe.index.dtype == "object":
            dframe.index = pd.to_datetime(dframe.index)

    # Remove duplicated column names. These will occur from libecl
    # when the user has repeated vector names in the summary SECTION
    dupes = dframe.columns.duplicated()
    if dupes.any():
        logger.warning(
            "Duplicated columns detected, check your DATA file "
            "for repeated vectors in the SUMMARY section"
        )
        logger.warning("Duplicates: %s", list(dframe.columns[dupes]))
        dframe = dframe.loc[:, ~dframe.columns.duplicated()]
    return dframe


def _df2pyarrow(dframe: pd.DataFrame) -> pyarrow.Table:
    """Construct a Pyarrow table from a dataframe, conserving metadata.

    All integer columns will have datatype int32, all floats will have float32
    as this is Eclipse specific.

    Metadata values will be written as strings. A True property is thus
    represented as the string "True" in the arrow object.

    The index in the dataframe is always assumed to be a time-index, but
    not necessarily a Pandas datetimetype (which is only of nanosecond precision).
    This index is always named DATE in the pyarrow table.
    """

    field_list: List[pyarrow.Field] = []
    field_list.append(pyarrow.field("DATE", pyarrow.timestamp("ms")))
    column_arrays = [dframe.index.to_numpy()]

    for colname in dframe.columns:
        if "meta" in dframe.attrs and colname in dframe.attrs["meta"]:
            # Boolean objects in the metadata dictionary must be converted to bytes:
            field_metadata = {
                bytes(key, encoding="ascii"): bytes(str(value), encoding="ascii")
                for key, value in dframe.attrs["meta"][colname].items()
            }
        else:
            field_metadata = {}
        if pd.api.types.is_integer_dtype(dframe.dtypes[colname]):
            dtype = pyarrow.int32()
        elif pd.api.types.is_string_dtype(dframe.dtypes[colname]):
            # Parameters are potentially merged into the dataframe.
            dtype = pyarrow.string()
        else:
            dtype = pyarrow.float32()
        field_list.append(pyarrow.field(colname, dtype, metadata=field_metadata))
        column_arrays.append(dframe[colname].to_numpy())

    schema = pyarrow.schema(field_list)

    return pyarrow.table(column_arrays, schema=schema)


def _merge_params(
    dframe: pd.DataFrame,
    paramfile: Optional[Union[str, Path]] = None,
    eclfiles: Union[str, EclFiles] = None,
) -> pd.DataFrame:
    """Locate parameters in a <key> <value> file and add to the dataframe.

    Will fetch parameters directly from a text file if provided, or look up
    the parameters.txt file based on the location of an Eclise run.
    """

    if paramfile is None and eclfiles is not None:
        param_files = parameters.find_parameter_files(eclfiles)
        logger.info("Loading parameters from files: %s", str(param_files))
        param_dict = parameters.load_all(param_files)
    elif (
        paramfile is not None
        and eclfiles is not None
        and not Path(paramfile).is_absolute()
    ):
        param_files = parameters.find_parameter_files(eclfiles, filebase=str(paramfile))
        logger.info("Loading parameters from files: %s", str(param_files))
        param_dict = parameters.load_all(param_files)
    elif paramfile is not None and Path(paramfile).is_absolute():
        logger.info("Loading parameters from file: %s", str(paramfile))
        param_dict = parameters.load(paramfile)
    else:
        raise ValueError("Not able to locate parameters.txt")
    logger.info("Loaded %d parameters", len(param_dict))
    for key in param_dict:
        # By converting to str we are more robust with respect to what objects are
        # read from the parameters.json/txt/yml. Since we are only going
        # to dump to csv, it should not cause side-effects that floats end up
        # as strings in the dataframe.
        dframe[key] = str(param_dict[key])
    return dframe


def smry_meta(eclfiles: EclFiles) -> Dict[str, Dict[str, Any]]:
    """Provide metadata for summary data vectors.

    A dictionary indexed by summary vector name is returned, and each
    value is dictionary with the metadata types provided by the underlying
    EclSum object:

    * unit (string)
    * is_total (bool)
    * is_rate (bool)
    * is_historical (bool)
    * get_num (int) (only provided if not None)
    * keyword (str)
    * wgname (str or None)
    """
    if isinstance(eclfiles, EclSum):
        eclsum = eclfiles
    else:
        eclsum = eclfiles.get_eclsum()

    meta: Dict[str, Dict[str, Any]] = {}
    for col in eclsum.keys():
        meta[col] = {}
        meta[col]["unit"] = eclsum.unit(col)
        meta[col]["is_total"] = eclsum.is_total(col)
        meta[col]["is_rate"] = eclsum.is_rate(col)
        meta[col]["is_historical"] = eclsum.smspec_node(col).is_historical()
        meta[col]["keyword"] = eclsum.smspec_node(col).keyword
        meta[col]["wgname"] = eclsum.smspec_node(col).wgname
        num = eclsum.smspec_node(col).get_num()
        if num is not None:
            meta[col]["get_num"] = num
    return meta


def _fix_dframe_for_libecl(dframe: pd.DataFrame) -> pd.DataFrame:
    """Fix a dataframe making it ready for EclSum.from_pandas()

    * Ensures that the index is always datetime, and sorted.
    * Removes BLOCK vectors, these are currently not supported as
      it requires knowledge of the grid dimensions. Warnings
      will be emitted for skipped columns

    Args:
        dframe: Dataframe to read. Will not be modified.

    Returns:
        Modified copy of incoming dataframe.
    """
    if dframe.empty:
        return dframe
    dframe = dframe.copy()
    if "DATE" in dframe.columns:
        # Infer datatype (Pandas cannot answer it) based on the first element:
        if isinstance(dframe["DATE"].values[0], str):
            # Do not use pd.Series.apply() here, Pandas would try to convert it to
            # datetime64[ns] which is limited at year 2262.
            dframe["DATE"] = pd.Series(
                [
                    dateutil.parser.parse(datestr)  # type: ignore
                    for datestr in dframe["DATE"]
                ],
                dtype="object",
                index=dframe.index,
            )
        if isinstance(dframe["DATE"].values[0], dt.date):
            dframe["DATE"] = pd.Series(
                [
                    dt.datetime.combine(dateobj, dt.datetime.min.time())
                    for dateobj in dframe["DATE"]
                ],
                dtype="object",
                index=dframe.index,
            )

        dframe.set_index("DATE", inplace=True)
    if not isinstance(
        dframe.index.values[0], (dt.datetime, np.datetime64, pd.Timestamp)
    ):
        raise ValueError(
            "dataframe must have a datetime index, got "
            f"{dframe.index.values[0]} of type {type(dframe.index.values[0])}"
        )
    dframe.sort_index(axis=0, inplace=True)

    # This column will appear if dataframes are naively written to CSV
    # files and read back in again.
    if "Unnamed: 0" in dframe:
        dframe.drop("Unnamed: 0", axis="columns", inplace=True)

    block_columns = [
        col for col in dframe.columns if (col.startswith("B") or col.startswith("LB"))
    ]
    if block_columns:
        dframe = dframe.drop(columns=block_columns)
        logger.warning(
            "Dropped columns with block data, not supported: %s",
            str({colname.partition(":")[0] + ":*" for colname in block_columns}),
        )

    return dframe


def df2eclsum(
    dframe: pd.DataFrame,
    casename: str = "SYNTHETIC",
) -> EclSum:
    """Convert a dataframe to an EclSum object

    Args:
        dframe: Dataframe with a DATE colum (or with the
            dates/datetimes in the index).
        casename: Name of Eclipse casename/basename to be used for the EclSum object
            If the EclSum object is later written to disk, this will be used
            to construct the filenames.
    """
    if dframe.empty:
        return None

    if casename.upper() != casename:
        raise ValueError(f"casename {casename} must be UPPER CASE")
    if "." in casename:
        raise ValueError(f"Do not use dots in casename {casename}")

    dframe = _fix_dframe_for_libecl(dframe)
    return _libecl_eclsum_from_pandas(casename, dframe)
    # return EclSum.from_pandas(casename, dframe)


def _libecl_eclsum_pandas_frame(
    eclsum: EclSum,
    time_index: Optional[Union[List[dt.date], List[dt.datetime]]] = None,
    column_keys: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Build a Pandas dataframe from an EclSum object.

    Temporarily copied from libecl to circumvent bug

    https://github.com/equinor/ecl/issues/802
    """
    if column_keys is None:
        keywords = EclSumKeyWordVector(eclsum, add_keywords=True)
    else:
        keywords = EclSumKeyWordVector(eclsum)
        for key in column_keys:
            keywords.add_keywords(key)

    # pylint: disable=protected-access
    if time_index is None:
        time_index = eclsum.dates  # Changed from libecl
        data = np.zeros([len(time_index), len(keywords)])
        EclSum._init_pandas_frame(
            eclsum, keywords, data.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        )
    else:
        time_points = eclsum._make_time_vector(time_index)
        data = np.zeros([len(time_points), len(keywords)])
        EclSum._init_pandas_frame_interp(
            eclsum,
            keywords,
            time_points,
            data.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        )

    # Do not give datetime64[ms] to Pandas, it will try to convert it
    # to datetime64[ns] and error hard if it is out of bounds (year 2262)
    assert isinstance(time_index[0], (dt.date, dt.datetime))
    frame = pd.DataFrame(
        index=time_index,
        columns=list(keywords),
        data=data,
    )

    # frame.index.type is now either datetime64[ns] or datetime.datetime (object)
    # depending on whether the date range ended before 2262.
    return frame


def _libecl_eclsum_from_pandas(
    case: str,
    frame: pd.DataFrame,
    dims: Optional[List[int]] = None,
    headers: Optional[List[tuple]] = None,
) -> EclSum:
    """Build an EclSum object from a Pandas dataframe.

    Temporarily copied from libecl to circumvent bug

    https://github.com/equinor/ecl/issues/802
    """
    start_time = frame.index[0]

    # Avoid Pandas or numpy timestamps, to avoid limitations
    # to timestamp64[ns] date boundaries (year 2262)
    if isinstance(start_time, pd.Timestamp):
        start_time = start_time.to_pydatetime()

    var_list = []
    # pylint: disable=protected-access
    if headers is None:
        header_list = EclSum._compile_headers_list(frame.columns.values, dims)
    else:
        header_list = EclSum._compile_headers_list(headers, dims)
    if dims is None:
        dims = [1, 1, 1]
    ecl_sum = EclSum.writer(case, start_time, dims[0], dims[1], dims[2])
    for keyword, wgname, num, unit in header_list:
        var_list.append(
            ecl_sum.addVariable(keyword, wgname=wgname, num=num, unit=unit).getKey1()
        )

    for idx, time in enumerate(frame.index):
        days = (time - start_time).days
        t_step = ecl_sum.addTStep(idx + 1, days)
        for var in var_list:
            t_step[var] = frame.iloc[idx][var]
    return ecl_sum


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser: parser to fill with arguments
    """
    parser.add_argument(
        "DATAFILE",
        help="Name of Eclipse DATA file. " + "UNSMRY file must lie alongside.",
    )
    parser.add_argument(
        "--time_index",
        type=str,
        help=(
            "Time resolution mnemonic; raw, daily, monthly or yearly. "
            "Data at a given point in time applies until the next point in time. "
            "If not raw, data will be interpolated. Use interpolated rate vectors "
            "with care. Default is raw, which will include clock times. first and last "
            "are also accepted and will print data for the first or the last date. "
        ),
        default="raw",
    )
    parser.add_argument(
        "--column_keys",
        nargs="+",
        help=(
            "Summary column vector wildcards, space-separated. "
            "Default is to include all summary vectors available."
        ),
    )
    parser.add_argument(
        "--start_date",
        type=str,
        help=(
            "Start at a specific date, in ISO format YYYY-MM-DD. "
            "Ignored if time_index is first or last"
        ),
        default="",
    )

    parser.add_argument(
        "--end_date",
        type=str,
        help=(
            "End at a specific date, in ISO format YYYY-MM-DD. "
            "Ignored if time_index is first or last"
        ),
        default="",
    )
    parser.add_argument(
        "-p",
        "--params",
        action="store_true",
        help="Merge key-value data from parameter file into each row.",
    )
    parser.add_argument(
        "--paramfile",
        type=str,
        help=(
            "Filename of key-value parameter file to look for if -p is set, "
            "relative to Eclipse DATA file or an absolute filename. "
            "If not supplied, parameters.{json,yml,txt} in "
            "{., .. and ../..} will be merged in."
        ),
        default=None,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help=(
            "Name of output file. Use '-' to write to stdout. " "Default 'summary.csv'"
        ),
        default="summary.csv",
    )
    parser.add_argument("--arrow", action="store_true", help="Write to pyarrow format")

    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def fill_reverse_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Fill a parser for the operation:  dataframe -> eclsum files"""

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Basename for Eclipse output files",
        default="SYNTSMRY",
    )
    parser.add_argument("csvfile", help="Name of CSV file with summary data.")
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    parser.add_argument("--debug", action="store_true", help="Be verbose")
    return parser


def summary_main(args) -> None:
    """Read summary data from disk and write CSV back to disk"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    eclbase = (
        args.DATAFILE.replace(".DATA", "").replace(".UNSMRY", "").replace(".SMSPEC", "")
    )

    eclfiles = EclFiles(eclbase)
    sum_df = df(
        eclfiles,
        time_index=args.time_index,
        column_keys=args.column_keys,
        start_date=args.start_date,
        end_date=args.end_date,
        params=args.params,
        paramfile=args.paramfile,
        datetime=False,
    )
    if args.arrow:
        sum_df = _df2pyarrow(sum_df)

    write_dframe_stdout_file(sum_df, args.output, index=True, caller_logger=logger)


def summary_reverse_main(args) -> None:
    """Entry point for usage with "csv2ecl summary" on the command line"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )

    summary_df = pd.read_csv(args.csvfile)
    logger.info("Parsed %s", args.csvfile)

    outputdir = Path(args.output).parent
    eclbase = Path(args.output).name

    # EclSum.fwrite() can only write to current directory:
    cwd = os.getcwd()
    eclsum = df2eclsum(summary_df, eclbase)
    try:
        os.chdir(outputdir)
        EclSum.fwrite(eclsum)
    finally:
        os.chdir(cwd)

    logger.info("Wrote to %s and %s", args.output + ".UNSMRY", args.output + ".SMSPEC")



"""Test module for satfunc2df"""
import re

import pytest

from ecl2df import inferdims

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed, nothing relevant in here then",
        allow_module_level=True,
    )


def test_injectsatnumcount():
    """Test that we always get out a string with TABDIMS"""
    assert "TABDIMS" in inferdims.inject_dimcount("", "TABDIMS", 0, 1)
    assert "TABDIMS" in inferdims.inject_dimcount("", "TABDIMS", 0, 1)
    assert "TABDIMS" in inferdims.inject_dimcount("TABDIMS", "TABDIMS", 0, 1)
    assert "99" in inferdims.inject_dimcount("", "TABDIMS", 0, 99)

    assert " 1* " in inferdims.inject_dimcount("", "TABDIMS", 1, 99)
    assert "*" not in inferdims.inject_dimcount("", "TABDIMS", 0, 99)

    assert "EQLDIMS" in inferdims.inject_dimcount("", "EQLDIMS", 0, 1)
    assert "EQLDIMS" in inferdims.inject_dimcount("", "EQLDIMS", 0, 1)
    assert "EQLDIMS" in inferdims.inject_dimcount("EQLDIMS", "EQLDIMS", 0, 1)
    assert "99" in inferdims.inject_dimcount("", "TABDIMS", 0, 99)


def test_guess_ntequil():
    """Test inferring the correct NTEQUIL"""
    assert inferdims.guess_dim("EQUIL\n200 2000/\n2000 2000/\n", "EQLDIMS", 0) == 2
    assert inferdims.guess_dim("EQUIL\n200 2000/\n", "EQLDIMS", 0) == 1
    assert inferdims.guess_dim("EQUIL\n200 2000 333/\n0 0/\n1 1/\n", "EQLDIMS", 0) == 3


def test_guess_satnumcount():
    """Test that we are able to guess the SATUM count in difficult cases"""
    # We always require a newline after a "/" in the Eclipse syntax
    # (anything between a / and \n is ignored)
    assert inferdims.guess_dim("SWOF\n0/\n0/\n", "TABDIMS", 0) == 2
    assert inferdims.guess_dim("SWOF\n0/\n0/ \n0/\n", "TABDIMS", 0) == 3
    assert inferdims.guess_dim("SWFN\n0/\n\n0/\n", "TABDIMS", 0) == 2
    assert inferdims.guess_dim("SGOF\n0/\n", "TABDIMS", 0) == 1
    assert inferdims.guess_dim("SGOF\n0/\n0/\n", "TABDIMS", 0) == 2
    assert inferdims.guess_dim("SGOF\n0/\n0/\n0/\n", "TABDIMS", 0) == 3
    assert (
        inferdims.guess_dim("SGOF\n0 0 0 0/\n0 0 0 0/\n0 0 0 0/\n", "TABDIMS", 0) == 3
    )
    assert (
        inferdims.guess_dim(
            "SGOF\n0 0 0 0 1 1 1 1/\n0 0 0 0 1 1 1 1/\n0 0 0 0 1 1 1/\n", "TABDIMS", 0
        )
        == 3
    )


def test_guess_dim():
    """Test error conditions"""
    with pytest.raises(ValueError, match="Only supports TABDIMS and EQLDIMS"):
        inferdims.guess_dim("SWOF\n0/\n0/\n", "WELLDIMS", 0)

    with pytest.raises(ValueError, match="Only support item 0 and 1 in TABDIMS"):
        inferdims.guess_dim("SWOF\n0/\n0/\n", "TABDIMS", 2)

    with pytest.raises(ValueError, match="Only item 0 in EQLDIMS can be estimated"):
        inferdims.guess_dim("EQUIL\n0/\n0/\n", "EQLDIMS", 1)


def test_inject_dimcount():
    """Test error conditions"""
    with pytest.raises(ValueError, match="Only supports TABDIMS and EQLDIMS"):
        inferdims.inject_dimcount("SWOF\n0/\n0/\n", "WELLDIMS", 0, 1)

    with pytest.raises(ValueError, match="Only support item 0 and 1 in TABDIMS"):
        inferdims.inject_dimcount("SWOF\n0/\n0/\n", "TABDIMS", 2, 1)

    with pytest.raises(ValueError, match="Only item 0 in EQLDIMS can be injected"):
        inferdims.inject_dimcount("EQUIL\n0/\n0/\n", "EQLDIMS", 1, 1)

    with pytest.raises(AssertionError, match="dimvalue must be larger than zero"):
        inferdims.inject_dimcount("SWOF\n0/\n0/\n", "TABDIMS", 1, 0, 0)


def test_inject_xxxdims_ntxxx():
    """Test the wrapper of inject_dimcount()"""

    with pytest.raises(AssertionError):
        inferdims.inject_xxxdims_ntxxx("WELLDIMS", "NTSFUN", "SWOF\n0/\n0/\n")
    with pytest.raises(AssertionError):
        inferdims.inject_xxxdims_ntxxx("TABDIMS", "NTFUN", "SWOF\n0/\n0/\n")

    # Repeated calls should be ok:
    deck_with_injection = inferdims.inject_xxxdims_ntxxx(
        "TABDIMS", "NTSFUN", "SWOF\n0/\n0/\n", 1
    )

    assert "TABDIMS\n 1 /" in re.sub(
        " +",
        " ",
        str(
            inferdims.inject_xxxdims_ntxxx(
                "TABDIMS", "NTSFUN", deck_with_injection, None
            )
        ),
    )
    # If no number is supplied, return the deck untouched:
    assert "TABDIMS\n 1 /" in re.sub(
        " +",
        " ",
        str(
            inferdims.inject_xxxdims_ntxxx(
                "TABDIMS", "NTSFUN", deck_with_injection, None
            )
        ),
    )
